
#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

from holado.common.context.session_context import SessionContext
import logging
from holado_core.common.resource.table_data_manager import TableDataManager
from holado_python.common.tools.datetime import DateTime, TIMEZONE_LOCAL
import os
from holado_system.system.filesystem.file import File
import re
from holado_core.common.exceptions.technical_exception import TechnicalException
from holado.holado_config import Config
from holado_xml.xml.stream_xml_file import StreamXMLFile
import json
from holado_core.common.tools.tools import Tools
import sys


logger = logging.getLogger(__name__)



class CampaignManager(object):
    """ Manage all campaigns
    """
    
    def __init__(self, db_name="campaigns"):
        super().__init__()
        
        self.__db_name = db_name
        self.__resource_manager = None
        
        self.__campaigns_table_manager = TableDataManager('campaign', 'campaigns', self.__get_campaigns_table_sql_create(), db_name=self.__db_name)
        self.__campaign_scenarios_table_manager = TableDataManager('campaign scenario', 'campaign_scenarios', self.__get_campaign_scenarios_table_sql_create(), db_name=self.__db_name)
    
    def initialize(self, resource_manager):
        self.__resource_manager = resource_manager
        
        self.__campaigns_table_manager.initialize(resource_manager)
        self.__campaigns_table_manager.ensure_db_exists()
        self.__campaign_scenarios_table_manager.initialize(resource_manager)
        self.__campaign_scenarios_table_manager.ensure_db_exists()
    
    def __get_db_client(self):
        return self.__resource_manager.get_db_client(self.__db_name)
    
    def __get_campaigns_table_sql_create(self):
        return """CREATE TABLE campaigns (
                id INTEGER PRIMARY KEY,
                name TEXT NOT NULL,
                report_path TEXT NOT NULL
            )"""

    def __get_campaign_scenarios_table_sql_create(self):
        return """CREATE TABLE campaign_scenarios (
                id INTEGER PRIMARY KEY,
                campaign_id INTEGER NOT NULL,
                name TEXT NOT NULL,
                report_path TEXT,
                status TEXT,
                status_at TEXT,
                details TEXT
            )"""

    def update_stored_campaigns(self):
        """ Update stored reports in DB with new campaigns
        """
        # Get report paths of campaigns to import
        dt_last_camp = self.__get_last_campaign_scenario_status_datetime()
        report_paths = self.__get_campaigns_report_paths(since_datetime=dt_last_camp)
        
        # Sort reports in time order
        report_paths = sorted(report_paths, key=lambda p: os.path.getmtime(os.path.join(p, 'report_summary_scenario_all.txt')))
        logger.info(f"reports to import: {report_paths}", msg_size_limit=None)
        
        # Import reports
        for report_path in report_paths:
            self.import_campaign_reports(report_path)
    
    def __get_campaigns_report_paths(self, since_datetime):
        reports_path = SessionContext.instance().path_manager.get_reports_path(name="test_runner", with_application_group=False)
        file_paths = SessionContext.instance().path_manager.find_files(reports_path, subdir_relative_path='report_summary_scenario_all.txt', since_datetime=since_datetime)
        return [os.path.dirname(p) for p in file_paths]
    
    def __get_last_campaign_scenario_status_datetime(self):
        """ From stored campaigns, return the datetime of the last scenario with an execution status
        """
        client = self.__get_db_client()
        
        query_str = f'''
            SELECT status_at
            FROM campaign_scenarios
            ORDER BY status_at DESC 
            LIMIT 1
        '''
        res_dict_list = client.execute(query_str, result_as_dict_list=True, as_generator=False)
        
        status_dt_str = res_dict_list[0]['status_at'] if res_dict_list else None
        status_dt = DateTime.str_2_datetime(status_dt_str, tz=TIMEZONE_LOCAL) if status_dt_str else None
        return status_dt
        
    def import_campaign_reports(self, report_path):
        """ Import reports of a campaign
        @param report_path Path to the campaign report
        """
        logger.info(f"Import campaign report '{report_path}'")
        
        # Add campaign
        camp_name = os.path.basename(report_path)
        camp_id = self.add_campaign_if_needed(camp_name, report_path)
        
        # Import scenario status
        try:
            # Import information from 'report_detailed_scenario.xml'
            self.__import_campaign_report_detailed_scenario(report_path, camp_id)
        except Exception as exc:
            logger.debug(f"Failed to import campaign report '{report_path}' from '{Config.campaign_manager_import_report_name}': {Tools.represent_exception(exc)}")  # @UndefinedVariable
            if not 'No such file or directory' in str(exc):
                sys.exit(1)
            # Try to import information from 'report_summary_scenario_all.txt'
            try:
                self.__import_campaign_report_summary_scenario_all(report_path, camp_id)
            except:
                raise TechnicalException(f"Failed to import campaign report '{report_path}'")
            else:
                logger.debug(f"Imported campaign report '{report_path}' from 'report_summary_scenario_all.txt'")
        else:
            logger.debug(f"Imported campaign report '{report_path}' from '{Config.campaign_manager_import_report_name}'")  # @UndefinedVariable
    
    def get_scenario_history(self, scenario_name=None, size=None):
        client = self.__get_db_client()
        placeholder = client._get_sql_placeholder()
        
        # Get data from DB
        where_clause = ""
        where_data = []
        if scenario_name is not None:
            where_clause = f"where name = {placeholder}"
            where_data.append(scenario_name)
        
        query_str = f'''
            SELECT *
            FROM campaign_scenarios
            {where_clause}
            ORDER BY name, status_at DESC 
        '''
        camp_scenarios_gen = client.execute(query_str, *where_data, result_as_dict_list=True, as_generator=True)
        
        # Build result
        res = []
        cur_scenario_name = None
        cur_scenario_statuses = None
        for cs in camp_scenarios_gen:
            # Manage new scenario
            if cur_scenario_name is None or cur_scenario_name != cs['name']:
                cur_scenario_statuses = []
                cur_scenario_name = cs['name']
                res.append({'name':cur_scenario_name, 'statuses':cur_scenario_statuses})
            
            # Add campaign info for this scenario execution if size limit is not reached
            if size is None or len(cur_scenario_statuses) < size:
                cur_scenario_statuses.append({'at':cs['status_at'], 'status':cs['status']})
            
        return res
        
    def add_campaign_if_needed(self, name, report_path):
        filter_data = {'report_path': report_path}
        if not self.__campaigns_table_manager.has_data(filter_data):
            self.__campaigns_table_manager.add_data(filter_data, {'name': name})
        camp = self.__campaigns_table_manager.get_data(filter_data)
        return camp['id']
    
    def update_or_add_campaign_scenario(self, campaign_id, name, *, report_path=None, status=None, status_at_str=None, details=None):
        filter_data = {'campaign_id': campaign_id, 'name': name}
        data = {}
        if report_path is not None:
            data['report_path'] = report_path
        if status is not None:
            data['status'] = status
        if status_at_str is not None:
            data['status_at'] = status_at_str
        if details is not None:
            data['details'] = details
        
        self.__campaign_scenarios_table_manager.update_or_add_data(filter_data, data)
        # camp_sce = self.__campaign_scenarios_table_manager.get_data(filter_data)
        # return camp_sce['id']
    
    def __import_campaign_report_detailed_scenario(self, report_path, camp_id):
        filepath = os.path.join(report_path, Config.campaign_manager_import_report_name)  # @UndefinedVariable
        with StreamXMLFile(filepath) as fin:
            scenarios = fin.read_elements_as_dict_list()
        if Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
            logger.trace(f"Elements read in '{filepath}':\n{scenarios}")
        
        regex_period = re.compile(r"^\[(.+) - ([^\]]+)\]$")
        for scenario_index, scenario in enumerate(scenarios):
            if 'scenario' not in scenario:
                raise TechnicalException(f"Unexpected structure of {scenario_index+1}'th scenario: {scenario}")
            scenario = scenario['scenario']
            
            name = scenario['file']
            scenario_report_path = scenario['report']
            status = scenario['validation_status']
            
            m = regex_period.match(scenario['scenario_period'])
            if m:
                status_at_str = m.group(2)
                # If report is done with dates in compact format, end datetime in period is not a valid datetime, use begin datetime instead
                if not DateTime.is_str_datetime(status_at_str[:10], '%Y-%m-%d'):
                    status_at_str = m.group(1)
            else:
                raise TechnicalException(f"Failed to import report of scenario '{name}' in campaign report file '{filepath}'")
            
            details = json.dumps( {k:scenario[k] for k in scenario.keys() if k not in ['file', 'scenario_period', 'report', 'validation_status']} )
            
            self.update_or_add_campaign_scenario(camp_id, name, report_path=scenario_report_path, status=status, status_at_str=status_at_str, details=details)
        
    def __import_campaign_report_summary_scenario_all(self, report_path, camp_id):
        file_path = os.path.join(report_path, 'report_summary_scenario_all.txt')
        lines = File(file_path, do_open=True, mode='rt').readlines(strip_newline=True)
        
        regex_line_period = re.compile(r"^\[(.+) - ([^\]]+)\] (.+?)(?: - .+)? - (.*)$")
        regex_line_time = re.compile(r"^(.+?) - (.+?)(?: - .+)? - (.+)$")
        regex_status = re.compile(r"^(.*?)(?: \(.*\)| => .*)?$")
        
        for line in lines:
            m = regex_line_period.match(line)
            if m:
                status_at_str = m.group(2)
                # If report is done with dates in compact format, end datetime in period is not a valid datetime, use begin datetime instead
                if not DateTime.is_str_datetime(status_at_str[:10], '%Y-%m-%d'):
                    status_at_str = m.group(1)
                scenario_name = m.group(3)
                status_info = m.group(4)
            else:
                m = regex_line_time.match(line)
                if not m:
                    raise TechnicalException(f"Unexpected line format in {file_path}: [{line}]")
                status_at_str = m.group(1)
                scenario_name = m.group(2)
                status_info = m.group(3)
            
            m = regex_status.match(status_info)
            status = m.group(1)
            
            self.update_or_add_campaign_scenario(camp_id, scenario_name, status=status, status_at_str=status_at_str)
        






#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

import logging
from holado_core.common.exceptions.technical_exception import TechnicalException
from holado_value.common.tables.value_table_manager import ValueTableManager
from holado_value.common.tables.converters.value_table_converter import ValueTableConverter
from typing import Union
from holado_core.common.tables.table_with_header import TableWithHeader
from holado_core.common.tools.string_tools import StrTools
from holado_binary.ipc.binary import Binary
from holado_core.common.tools.tools import Tools
from holado_python.standard_library.typing import Typing
import copy
from holado_core.common.exceptions.functional_exception import FunctionalException
from holado_ais.ais.enums import AISMessageType

logger = logging.getLogger(__name__)

try:
    import pyais
    from bitarray import bitarray
    from pyais.encode import ais_to_nmea_0183
    from pyais.messages import TagBlock
    with_pyais = True
    
    # Patch pyais with some fixes
    import holado_ais.ais.patch_pyais  # @UnusedImport
except Exception as exc:
    if Tools.do_log(logger, logging.DEBUG):
        logger.debug(f"AIS is not available. Initialization failed on error: {exc}")
    with_pyais = False
# else:
#     if Tools.do_log(logger, logging.DEBUG):
#         logger.debug(f"AIS is available !")


class AISMessages(object):
    
    @classmethod
    def is_available(cls):
        return with_pyais
    
    ###########################################################################
    # Methods depending to pyais
    ###########################################################################
    if with_pyais:
        
        def __init__(self):
            self.__tag_block_group_id = 0
        
        def default_message_data(self, msg_type:Union[int,str]):
            """Return default data values for required data by pyais but not functionnaly required.
            It is useful when generating custom messages, for example for tests.
            """
            return {}
        
        def new_message_as_bitarray_bytes(self, msg_type:Union[int,str], data:Union[dict,TableWithHeader]):
            msg = self.new_message(msg_type, data)
            res = self.convert_message_to_bytes(msg)
            return res
        
        def new_message(self, msg_type:Union[int,str], data:Union[dict,TableWithHeader]):
            if isinstance(data, TableWithHeader):
                data = self.__convert_table_to_dict(data)
            if not isinstance(data, dict):
                raise TechnicalException(f"Unexpected data type '{Typing.get_object_class_fullname(data)}'")
            
            # Add data key 'msg_type' if missing
            if 'msg_type' not in data:
                data['msg_type'] = self.__get_message_type_int(msg_type)
            
            message_type = self.__get_message_type(msg_type)
            res = message_type.create(**data)
            return res
        
        def decode_nmea_raw_as_dict(self, raw, merge_tag_block_and_message_dicts=False, enum_as_int=False, ignore_spare=True):
            # Convert NMEA raw to sentences
            sentences = self.split_raw_to_sentences(raw)
            
            # Convert sentences to message type as dict
            msg = self.decode(*sentences)
            msg_dict = msg.asdict(enum_as_int=enum_as_int, ignore_spare=ignore_spare)
            
            # Extract tagblock of first sentence if present, and add tagblock information in result
            tag_block_str, _ = self.split_sentence_to_tag_block_and_message(sentences[0])
            if tag_block_str is None:
                tag_block_dict = None
            else:
                if not isinstance(tag_block_str, bytes):
                    tag_block_str = tag_block_str.encode('utf8')
                tag_block = TagBlock(tag_block_str)
                tag_block.init()
                tag_block_dict = tag_block.asdict()
                tag_block_dict = {k:v for k,v in tag_block_dict.items() if k != 'raw' and v is not None}
            
            # Build result
            if merge_tag_block_and_message_dicts:
                res = msg_dict
                if tag_block_dict is not None:
                    res.update(tag_block_dict)
            elif tag_block_dict is not None:
                res = (tag_block_dict, msg_dict)
            else:
                res = msg_dict
            return res
            
        def decode(self, *args, encoded_msg=None, parts_separator=b'\n'):
            if encoded_msg is not None:
                if isinstance(encoded_msg, str):
                    encoded_bytes = encoded_msg.encode("utf-8")
                elif isinstance(encoded_msg, bytes):
                    encoded_bytes = encoded_msg
                else:
                    raise TechnicalException(f"Parameter 'encoded_msg' must be of type str or bytes")
                args = encoded_bytes.split(parts_separator)
            return pyais.decode(*args, error_if_checksum_invalid=True)
        
        def decode_message(self, msg_type:Union[int,str], bit_array:Union[str,bitarray]):
            if isinstance(bit_array, str):
                if StrTools.is_hex(bit_array) and not StrTools.is_bitarray(bit_array):
                    bit_array = Binary.convert_hex_str_to_bin_str(bit_array)
                bit_array = bitarray(bit_array)
            if not isinstance(bit_array, bitarray):
                raise TechnicalException(f"Bad parameter 'bit_array' (accepted types: bitarray, str)")
                
            message_type = self.__get_message_type(msg_type)
            return message_type.from_bitarray(bit_array)
        
        def encode_raw_payload(self, payload:Union[str,bytes], tag_block: TagBlock = None, with_tag_block=True, talker_id: str = "AIVDM", **kwargs):
            from pyais.util import encode_ascii_6
            
            if isinstance(payload, bytes):
                payload_bytes = payload
            elif isinstance(payload, str):
                payload_bytes = bytes.fromhex(payload)
            else:
                raise TechnicalException(f"Unmanaged payload type '{Typing.get_object_class_fullname(payload)}'")
            
            payload_bitarray = bitarray()
            payload_bitarray.frombytes(payload_bytes)
            payload_ascii6, fill_bits = encode_ascii_6(payload_bitarray)
            
            seq_id = kwargs.pop('seq_id') if 'seq_id' in kwargs else self.__tag_block_group_id % 10
            radio_channel = kwargs['radio_channel'] if 'radio_channel' in kwargs else 'A'
            group_id = kwargs.pop('group_id') if 'group_id' in kwargs else None
            
            # sentences = HA_ais_to_nmea_0183(payload_ascii6, talker_id, radio_channel, fill_bits, seq_id)
            sentences = ais_to_nmea_0183(payload_ascii6, talker_id, radio_channel, fill_bits, seq_id)
            return self.__encode_add_tag_block(sentences, tag_block, with_tag_block=with_tag_block, group_id=group_id)
            
        def encode(self, data, tag_block: TagBlock = None, with_tag_block=True, talker_id: str = "AIVDM", **kwargs):
            if isinstance(data, pyais.messages.Payload):
                return self.encode_msg(data, tag_block=tag_block, with_tag_block=with_tag_block, talker_id=talker_id, **kwargs)
            elif isinstance(data, dict):
                return self.encode_data(data, tag_block=tag_block, with_tag_block=with_tag_block, talker_id=talker_id, **kwargs)
            elif isinstance(data, TableWithHeader):
                return self.encode_table(data, tag_block=tag_block, with_tag_block=with_tag_block, talker_id=talker_id, **kwargs)
            else:
                raise TechnicalException(f"Unexpected data type '{Typing.get_object_class_fullname(data)}'")
        
        def encode_msg(self, msg, tag_block: TagBlock = None, with_tag_block=True, talker_id: str = "AIVDM", **kwargs):
            if not isinstance(msg, pyais.messages.Payload):
                raise TechnicalException(f"Parameter 'msg' is not an AIS message (obtained type: {Typing.get_object_class_fullname(msg)})")
            
            seq_id = kwargs.pop('seq_id') if 'seq_id' in kwargs else self.__tag_block_group_id % 10
            group_id = kwargs.pop('group_id') if 'group_id' in kwargs else None
            
            sentences = pyais.encode_msg(msg, seq_id=seq_id, talker_id=talker_id, **kwargs)
            return self.__encode_add_tag_block(sentences, tag_block, with_tag_block=with_tag_block, group_id=group_id)
        
        def encode_data(self, data, tag_block: TagBlock = None, with_tag_block=True, talker_id: str = "AIVDM", **kwargs):
            if not isinstance(data, dict):
                raise TechnicalException(f"Parameter 'data' is not a dict (obtained type: {Typing.get_object_class_fullname(data)})")
            
            seq_id = kwargs.pop('seq_id') if 'seq_id' in kwargs else self.__tag_block_group_id % 10
            group_id = kwargs.pop('group_id') if 'group_id' in kwargs else None
            
            sentences = pyais.encode_dict(data, seq_id=seq_id, talker_id=talker_id, **kwargs)
            return self.__encode_add_tag_block(sentences, tag_block, with_tag_block=with_tag_block, group_id=group_id)
        
        def encode_table(self, table, tag_block: TagBlock = None, with_tag_block=True, talker_id: str = "AIVDM", **kwargs):
            data = self.__convert_table_to_dict(table)
            return self.encode_data(data, tag_block=tag_block, with_tag_block=with_tag_block, talker_id=talker_id, **kwargs)
        
        def __encode_add_tag_block(self, encoded_sentences, tag_block: TagBlock = None, with_tag_block=True, group_id=None):
            from pyais.messages import TagBlockGroup
            
            if not with_tag_block:
                return encoded_sentences
            
            res = []
            nb_sentences = len(encoded_sentences)
            if nb_sentences > 1 and group_id is None:
                group_id = self.__tag_block_group_id
                # Increment group_id for next message needing a group_id
                self.__new_tag_block_group_id()
            elif nb_sentences == 1:
                # No group tag block is needed
                group_id = None
            
            for index, sentence in enumerate(encoded_sentences):
                sentence_tag_block = copy.deepcopy(tag_block) if index == 0 else None
                if group_id is not None:
                    if sentence_tag_block is None:
                        sentence_tag_block = TagBlock(None)
                    sentence_tag_block._group = TagBlockGroup(msg_id=index+1, total=nb_sentences, group_id=group_id)
                
                if sentence_tag_block is not None:
                    res.append('\\' + sentence_tag_block.to_raw().decode() + '\\' + sentence)
                else:
                    res.append(sentence)
            
            return res
        
        def convert_message_to_binary_str(self, msg):
            return msg.to_bitarray().to01()
        
        def convert_message_to_bytes(self, msg):
            return msg.to_bitarray().tobytes()
        
        def __get_message_type(self, msg_type:Union[int,str]):
            import importlib
            
            if isinstance(msg_type, str):
                msg_type = AISMessageType[msg_type].value
            
            module = importlib.import_module('pyais.messages')
            res = getattr(module, f"MessageType{msg_type}")
            return res
        
        def __get_message_type_int(self, msg_type:Union[int,str]):
            if isinstance(msg_type, int):
                return msg_type
            
            if isinstance(msg_type, str):
                msg_type = AISMessageType[msg_type]
            if not isinstance(msg_type, AISMessageType):
                raise TechnicalException(f"Unmanaged message type of class '{Typing.get_object_class_fullname(msg_type)}'")
            return msg_type.value
            
        def __convert_table_to_dict(self, table):
            if not ValueTableManager.is_table_with_header(table):
                raise TechnicalException(f"Parameter 'table' is not a table with header (obtained type: {Typing.get_object_class_fullname(table)})")
            
            if ValueTableManager.verify_table_is_name_value_table(table, raise_exception=False):
                res = ValueTableConverter.convert_name_value_table_2_dict(table)
            else:
                res = ValueTableConverter.convert_table_with_header_to_dict(table)
            return res
        
    
    
    ###########################################################################
    # Methods independent of pyais
    ###########################################################################
    
    def __new_tag_block_group_id(self):
        self.__tag_block_group_id += 1
        if self.__tag_block_group_id > 9999:
            self.__tag_block_group_id = 0
        return self.__tag_block_group_id
    
    def split_raw_to_sentences(self, ais_raw):
        if isinstance(ais_raw, str):
            split_char = '\n'
        elif isinstance(ais_raw, bytes):
            split_char = b'\n'
        else:
            raise FunctionalException(f"Raw must by of type str or bytes (obtained type: {Typing.get_object_class_fullname(ais_raw)})")
        
        res = ais_raw.split(split_char)
        
        return res
    
    def split_sentence_to_tag_block_and_message(self, ais_sentence):
        if isinstance(ais_sentence, str):
            split_char = '\\'
        elif isinstance(ais_sentence, bytes):
            split_char = b'\\'
        else:
            raise FunctionalException(f"Sentence must by of type str or bytes (obtained type: {Typing.get_object_class_fullname(ais_sentence)})")
        
        res = ais_sentence.split(split_char)
        if len(res) > 1:
            # Tag block is present, remove part before first '\'
            del res[0]
        else:
            # Tag block is not present, insert it as None
            res.insert(0, None)
            
        return res
    
    def convert_tag_block_to_dict(self, tag_block):
        if isinstance(tag_block, str):
            checksum_delimiter = '*'
            param_delimiter = ','
            name_delimiter = ':'
        elif isinstance(tag_block, bytes):
            checksum_delimiter = b'*'
            param_delimiter = b','
            name_delimiter = b':'
        else:
            raise FunctionalException(f"Tag Block must by of type str or bytes (obtained type: {Typing.get_object_class_fullname(tag_block)})")
        
        # Remove checksum
        try:
            chk_index = tag_block.index(checksum_delimiter)
        except:
            chk_index = None
        tb_str = tag_block[:chk_index] if chk_index else tag_block
        
        # Split parameters
        tb_list = tb_str.split(param_delimiter)
        
        # Extract parameter names and values
        param_list = [e.split(name_delimiter) for e in tb_list]
        res = {p[0]:p[1] for p in param_list}
        
        return res
    
    def split_message_to_fields(self, ais_message):
        split_char = self.__get_fields_delimiter_for_message(ais_message)
        res = ais_message.split(split_char)
        return res
    
    def __get_fields_delimiter_for_message(self, ais_message):
        if isinstance(ais_message, str):
            return ','
        elif isinstance(ais_message, bytes):
            return b','
        else:
            raise FunctionalException(f"Message must by of type str or bytes (obtained type: {Typing.get_object_class_fullname(ais_message)})")
        
    def replace_message_id_in_message(self, ais_message, replacement_id):
        if type(ais_message) != type(replacement_id):
            raise FunctionalException(f"Replacement ID must of same type than message ({Typing.get_object_class_fullname(ais_message)})")
        
        fields = self.split_message_to_fields(ais_message)
        fields[3] = replacement_id
        
        split_char = self.__get_fields_delimiter_for_message(ais_message)
        return split_char.join(fields)
        
        
    



#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of self software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and self permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

import logging
import abc
from holado_core.common.handlers.redo import Redo
from selenium.common.exceptions import WebDriverException,\
    StaleElementReferenceException
from holado_core.common.tools.tools import Tools

logger = logging.getLogger(__name__)


class SeleniumRedo(Redo):
    """ Redo on Selenium actions.
    """
    __metaclass__ = abc.ABCMeta

    def __init__(self, name, selenium_driver):
        super().__init__(name)
        
        self.__selenium_driver = selenium_driver
        
        # Manage iteration on scroll_by_y
        self.__with_scroll_by_y = False
        self.__scroll_by_y = 0
        
        # Ignore stale exception
        self.ignoring(StaleElementReferenceException)
        # self.ignoring(UnreachableBrowserException)
        
        # Reduce process timeout, in order to execute several retries
        self.with_process_timeout_for_retry(3)
        
        # Deactivate polling to reduce stale exceptions
        self.polling_every(0)
        
        # Manage selenium freeze
        self.with_nb_successive_failure_before_error(3)
        
        # Manage element that is under change (WebDriverException with message "unknown error")
        self.ignoring(WebDriverException)
    
    @property
    def internal_api(self):
        return self.__selenium_driver.internal_api
    
    @property
    def scroll_by_y(self):
        return self.__scroll_by_y
    
    def with_iteration_on_scroll_by_y(self):
        """
        Activate detection of not clickable element exception and iterate on scroll_by_y
        @return self
        """
        self.__with_scroll_by_y = True
        return self
    
    def _execute_after_ignored(self, exception):
        # If element is not clickable, increase scroll by y after scroll_into_visible
        if self.__with_scroll_by_y and isinstance(exception, WebDriverException) and " is not clickable at point " in str(exception):
            old_scroll_by_y = self.__scroll_by_y
            self.__scroll_by_y = self.__scroll_by_y * 5 // 4 - 10
            logger.warning(f"Element was not clickable ; update scroll_by_y : {old_scroll_by_y} -> {self.__scroll_by_y}")
            return
        
        # Only ignore WebDriverException containing "unknown error" in message
        if type(exception) == WebDriverException and "unknown error" not in str(exception):
            Tools.raise_same_exception_type(exception, str(exception))
    
    
    
    
    
    



import os
import sys
import argparse
import logging

here = os.path.abspath(os.path.dirname(__file__))

holado_path = os.path.join(here, "..", "..")
sys.path.insert(0, os.path.join(holado_path, "src"))


from holado_core.common.exceptions.functional_exception import FunctionalException
from holado_report.report.analyze.scenario_duration_manager import ScenarioDurationManager



def __main(args):
    if not os.path.exists(args.input):
        raise FunctionalException(f"Input file '{args.input}' doesn't exist")
    if not os.path.exists(args.features):
        raise FunctionalException(f"Features folder '{args.features}' doesn't exist")
    
    features_path = os.path.abspath(args.features)
    
    sdm = ScenarioDurationManager()
    sdm.update_scenario_durations(args.input, features_path, args.tag_prefix, args.dry_run)
    
    logging.info(f"Finished")
    

if __name__ == "__main__":
    # Arguments
    descr = """Update scenario files by adding/updating the scenario duration tag.
    It needs as input the file created with script generate_scenario_duration_tags.py with '-a' option.
    
    Notes: 
        * The intermediate file between the two scripts exists to let the team the possibility to validate the automatic modifications, 
          by modifying the content of the column 'Apply' that indicates to this script if the modification has to be applied automatically or not.
        * Team can also choose another new tag by modifying the content of the column 'New Tag'. 
          It is recommended in this case to suffix the tag by "!fixed". 
          The suffix "!fixed" is recognized by the scripts as a tag fixed manually that, by default, shouldn't be overridden by automatic new tag.
          
    Example 1: Simulate the modifications the script would apply
            
            python update_scenario_duration_tags.py -i {path_to_scenario_duration_tags_file} -f {path_to_features} -d
            
    Example 2: Apply modifications 
            
            python update_scenario_duration_tags.py -i {path_to_scenario_duration_tags_file} -f {path_to_features}
    """
    
            
    parser = argparse.ArgumentParser(description=descr, formatter_class=argparse.RawTextHelpFormatter)

    parser.add_argument('-i', '--input', dest='input', required=True,
                   help='Path to the test report folder.')
    
    parser.add_argument('-f', '--features', dest='features', required=True,
                   help='Path to the features folder.')
    
    parser.add_argument('-tp', '--tag-prefix', dest='tag_prefix', default="ScenarioDuration=", 
                   help="Prefix of the duration tag in .feature files.")
    
    parser.add_argument('-d', '--dry-run', dest='dry_run', default=False, action="store_true",
                   help='If set, no scenario file is updated, it just logs updates it would have done.')

    args = parser.parse_args()

    logging.basicConfig(format = '{asctime:s} | {thread:-5d} | {levelname:5s} | {module:35s} | {message:s}', style = '{', level=logging.DEBUG, handlers=[logging.StreamHandler()])
    
    # Execute
    try:
        __main(args)
    except FunctionalException as exc:
        logging.error(f"Update is interrupted on error: {exc}")
        sys.exit(1)
    except:
        logging.exception("An error occured")
        sys.exit(1)

    sys.exit(0)
    
    



#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

import logging
from holado_report.report.builders.report_builder import ReportBuilder
from holado_system.system.filesystem.file import File
from holado_core.common.exceptions.technical_exception import TechnicalException

logger = logging.getLogger(__name__)



class SummaryByCategoryReportBuilder(ReportBuilder):
    def __init__(self, filepath, file_format='txt', exclude_statuses=None, exclude_categories=None):
        self.__file_format = file_format.lower()
        self.__exclude_statuses = exclude_statuses
        self.__exclude_categories = exclude_categories
        
        if self.__file_format == 'txt':
            self.__file = File(filepath, mode='wt')
        else:
            raise TechnicalException(f"Unmanaged format '{self.__file_format}' (possible formats: 'txt')")
        
        self.__counter_by_category = {}
        self.__categories_order = [
                'Regression',
                'Always Failed',
                'Random',
                'Regression but Not Relevant',
                'Always Not Relevant',
                'Random but Not Relevant',
                'Always Success',
                'Fixed',
                'Unknown'
            ]
        
    def build(self):
        '''
        The file is built after each scenario
        '''
        pass
        
    def after_scenario(self, scenario, scenario_report=None):
        from holado_report.report.report_manager import ReportManager
        category_validation, status_validation, _, _, _, _ = ReportManager.get_current_scenario_status_information(scenario)
        if category_validation is not None:
            ind = category_validation.find(' (')
            category = category_validation[:ind] if ind > 0 else category_validation
        else:
            category = None
        
        # Manage excluded scenarios
        if self.__exclude_statuses and status_validation in self.__exclude_statuses:
            return
        if category is not None and self.__exclude_categories and category in self.__exclude_categories:
            return
        
        if category is not None:
            if category not in self.__counter_by_category:
                self.__counter_by_category[category] = 0
            self.__counter_by_category[category] += 1
            
            # Update categories order with unexpected category
            if category not in self.__categories_order:
                self.__categories_order.append(category)
            
            self.__update_file()
        
    def __update_file(self):
        list_by_group = {
            'Success': [(k,self.__counter_by_category[k]) for k in self.__categories_order 
                        if k in self.__counter_by_category and k in ['Always Success', 'Fixed']],
            'Failed': [(k,self.__counter_by_category[k]) for k in self.__categories_order 
                       if k in self.__counter_by_category and k in ['Regression', 'Always Failed', 'Random']],
            'Not Relevant': [(k,self.__counter_by_category[k]) for k in self.__categories_order 
                             if k in self.__counter_by_category and k in ['Regression but Not Relevant', 'Always Not Relevant', 'Random but Not Relevant']]
            }
        others_list = [(k,v) for k,v in self.__counter_by_category.items() 
                       if k not in ['Always Success', 'Fixed', 'Regression', 'Always Failed', 'Random', 'Regression but Not Relevant', 'Always Not Relevant', 'Random but Not Relevant']]
        
        lines = []
        for group_name, group_list in list_by_group.items():
            if group_list:
                lines.append(f"{group_name:>12s}: {sum([x[1] for x in group_list]):5d}  ({' ; '.join([f'{x[0]}: {x[1]}' for x in group_list])})")
            else:
                lines.append(f"{group_name:>12s}: {0:5d}")
        if others_list:
            lines.append(f"{'Others':>12s}: {sum([x[1] for x in others_list]):5d}  ({' ; '.join([f'{x[0]}: {x[1]}' for x in others_list])})")

        with self.__file as fout:
            fout.writelines(lines)
        
        




#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

from holado_docker.sdk.docker.docker_client import DockerClient
from holado_core.common.exceptions.technical_exception import TechnicalException
import logging
from holado_core.tools.abstracts.service import Service
from holado_system.system.command.command import CommandStates
from holado_core.common.tools.tools import Tools

logger = logging.getLogger(__name__)


class DockerService(Service):
    @classmethod
    def is_available(cls):
        return DockerClient.is_available()
    
    def __init__(self, name):
        super().__init__(name) 
        self.__docker_client = None
        self.__docker_container = None
    
    @property
    def status(self):
        if self.__docker_container is not None and self.__docker_client.has_container(self.name):
            status = self.__docker_container.status
            if status in ["running", "paused", "restarting"]:
                return CommandStates.Running
            elif status == "exited":
                result = self.__docker_container.wait(timeout=10)
                error_code = result["StatusCode"]
                if error_code == 0:
                    return CommandStates.Success
                else:
                    return CommandStates.Error
            else:
                raise TechnicalException(f"Unmanaged docker status '{status}'")
        else:
            return None
    
    def run_as_docker(self, image, remove_existing=True, remove=True, auto_stop=True, **kwargs):
        """
        Run the service as a docker image.
        
        :param image: Image name
        :param remove_existing: If true (default) and docker already exists, remove docker before creating a new one 
        :param remove: If true (default), remove the docker at stop
        :param auto_stop: If true (default), automatically stop docker when docker instance is released
        """ 
        self.__docker_client = DockerClient()
        self.__docker_container = self.__docker_client.run_container(self.name, image, remove_existing=remove_existing, remove=remove, auto_stop=auto_stop, **kwargs)

    def stop(self):
        if Tools.do_log(logger, logging.DEBUG):
            logger.debug(f"Stopping service '{self.name}'")
        if self.__docker_container is not None:
            if self.__docker_client.has_container(self.name) and self.__docker_client.get_container(self.name).status == "running":
                self.__docker_client.stop_container(self.name)
        else:
            raise TechnicalException(f"Service '{self.name}' is not started")




#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

import logging
from holado.common.handlers.undefined import default_value

logger = logging.getLogger(__name__)


class HALogger(logging.Logger):
    default_message_size_limit = None
    
    def __init__(self, name, level=logging.NOTSET):
        super().__init__(name, level)
    
    def _log(self, level, msg, args, exc_info=None, extra=None, stack_info=False, stacklevel=1, msg_size_limit=default_value):
        # def find_handlers(logger_):
        #     index = 0
        #     while len(logger_.handlers) == 0:
        #         if logger_.parent is not None:
        #             logger_ = logger_.parent
        #             index += 1
        #         else:
        #             break
        #     return index, logger_.handlers
        # print(f"-------------- {id(self)} | {self.name} | {level}/{self.getEffectiveLevel()} | {msg} | {find_handlers(self)=} | {self.filters=}")
        try:
            from holado_core.common.tools.tools import Tools
        except ImportError as exc:
            if "Python is likely shutting down" in str(exc):
                return
            else:
                raise exc
        
        if msg_size_limit is default_value:
            msg_size_limit = HALogger.default_message_size_limit
        msg_to_log = Tools.truncate_text(msg, msg_size_limit)
            
        logging.Logger._log(self, level, msg_to_log, args, exc_info=exc_info, extra=extra, stack_info=stack_info, stacklevel=stacklevel)
    
    # def setLevel(self, level)->None:
    #     logger.print(f"Change logger {self} level to {level}")
    #     logging.Logger.setLevel(self, level)
    
class HARootLogger(HALogger):
    """
    Implementation is a copy of logging.RootLogger
    """
    def __init__(self, level):
        """
        Initialize the logger with the name "root".
        """
        super().__init__("root", level)

    def __reduce__(self):
        return logging.getLogger, ()




    



#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

import logging
from holado.holado_config import Config
from holado_core.common.tools.tools import Tools

logger = logging.getLogger(__name__)



class DynamicTextManager(object):
    """
    @summary: Dynamic text manager.
    For every text it manages, it modifies the texts by adding a dynamic suffix.
    The suffix is computed from the timestamp of the moment when the text is registered to be dynamic.
    
    The scope of the text can be specified, so that the same text can have a different suffix by scope.
    Usual usage is to use the thread id as scope, so that the same text has a different suffix by thread.
    The default scope is None.
    """
    
    def __init__(self, name):
        self.__name = name
        self.__texts = {}
        self.__unique_value_manager = None

    def initialize(self, unique_value_manager):
        self.__unique_value_manager = unique_value_manager
        
    def __set_text(self, scope, key, value):
        if scope not in self.__texts:
            self.__texts[scope] = {}
        self.__texts[scope][key] = value
        
    def has(self, text, scope=None):
        return self.__has_text(text, scope=scope)
    
    def get(self, text, is_dynamic=True, scope=None):
        if is_dynamic:
            return self.__get_text(text, scope=scope)
        else:
            return text
        
    def __get_text(self, text, scope=None):
        if not self.__has_text(text, scope=scope):
            self.__new_text(text, scope=scope)
        return self.__texts[scope][text]
        
    def __has_text(self, text, scope=None):
        return scope in self.__texts and text in self.__texts[scope]
    
    def __new_text(self, text, scope=None):
        # Compute text with unique suffix
        suffix = self.__unique_value_manager.new_string(padding_length=Config.unique_string_padding_length)
        res = text + suffix

        # Store dynamic text
        self.__set_text(scope, text, res)
        
        if Tools.do_log(logger, logging.DEBUG):
            logger.debug(f"[{self.__name}] New dynamic text in scope {scope}: [{text}] -> [{res}]")




#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of self software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and self permission notice shall be included in all copies or substantial portions of the Software.
#
# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

import logging
from holado_ui_selenium.ui.gui.selenium.inspectors.selenium_inspector import SeleniumInspector
from holado_ui.ui.inspectors.tools.ui_inspect_context import UIInspectContext
from holado_core.common.inspectors.tools.inspect_parameters import InspectParameters
from holado_ui_selenium.ui.gui.selenium.criterias.selenium_criteria import SeleniumCriteria
from holado_core.common.criterias.or_criteria import OrCriteria
from holado_core.common.finders.else_finder import ElseFinder
import re
from holado_core.common.finders.then_finder import ThenFinder
from holado_core.common.finders.tools.enums import FindType
from holado_ui_selenium.ui.gui.selenium.criterias.selenium_criteria_and_verifier import SeleniumCriteriaAndVerifier
from holado_core.common.exceptions.technical_exception import TechnicalException
from holado.common.context.session_context import SessionContext
from holado_ui_selenium.ui.gui.selenium.finders.by_selenium_finder import BySeleniumFinder
from holado_core.common.criterias.and_criteria import AndCriteria
from holado_core.common.tools.tools import Tools

logger = logging.getLogger(__name__)



class HtmlSeleniumInspector(SeleniumInspector):
    """ Selenium inspector with Angular support.
    """
    
    def __init__(self):
        super().__init__("html")
    
    @SeleniumInspector.default_inspect_builder.getter  # @UndefinedVariable
    def default_inspect_builder(self):
        res = super().default_inspect_builder
    
        res.default_parameters.add_finder_types([
            "label", "label as div", "label as span",
            "select",
            "text-node", "text-content"
            ])
        
        return res
    
    def _get_criteria_editable_element(self):
        res = OrCriteria()
        
        criteria = SeleniumCriteria(self)
        criteria.tag_name = "input"
        criteria.set_attribute_pattern("type", "text|password")
        res.add_criteria(criteria)
        
        # Criteria for select
        criteria = SeleniumCriteria(self)
        criteria.tag_name = "select"
        res.add_criteria(criteria)
        
        # Criteria for textarea
        criteria = SeleniumCriteria(self)
        criteria.tag_name = "textarea"
        res.add_criteria(criteria)

        return res
    
    def _get_criteria_text_element(self):
        res = OrCriteria()
        
        # Editable form
        res.add_criteria( self._get_criteria_editable_element() )
        
        # Element containing text
        criteria = SeleniumCriteria(self)
        criteria.tag_name_pattern = "div|td|span|p"
#        criteria.content_text_pattern = "[^\\s]+";    # With contentText, a text element containing another text element won't be validated
        criteria.inner_text_pattern = "[^\\s]+"
        res.add_criteria(criteria)
        
        return res
    
    def get_finder_label(self, label, inspect_context:UIInspectContext=None, inspect_parameters:InspectParameters=None):
        if inspect_context is None:
            inspect_context = self.inspect_builder.context()
        
        return self._build_result_finder_from_list([
                # Add search by label
                self.get_finder_by_tag_and_inner_text("label", label, inspect_context.without_ui_context().with_finder_type("label"), inspect_parameters),
                # Add search by div
                self.get_finder_by_tag_and_inner_text("div", label, inspect_context.without_ui_context().with_finder_type("label as div"), inspect_parameters),
                # Add search by span
                self.get_finder_by_tag_and_inner_text("span", label, inspect_context.without_ui_context().with_finder_type("label as span"), inspect_parameters)
            ], inspect_context, inspect_parameters)
        return self.get_finder_by_tag("select", inspect_context.with_finder_type("select"), inspect_parameters)
    
    def get_finder_select(self, inspect_context:UIInspectContext=None, inspect_parameters:InspectParameters=None):
        if inspect_context is None:
            inspect_context = self.inspect_builder.context()
        
        return self.get_finder_by_tag("select", inspect_context.with_finder_type("select"), inspect_parameters)
    
    def get_finder_input_text(self, inspect_context:UIInspectContext=None, inspect_parameters:InspectParameters=None):
        """
        @param inspect_context Inspect context
        @param inspect_parameters Inspect parameters
        @return Finder for input text
        """
        criteria = SeleniumCriteria(self)
        criteria.tag_name = "input"
        criteria.set_attribute("type", "text")
        
        res = BySeleniumFinder(criteria=criteria, description="text input")
        return self._build_result_finder(res, inspect_context, inspect_parameters)
    
    def get_finder_table(self, inspect_context:UIInspectContext=None, inspect_parameters:InspectParameters=None):
        return self.get_finder_by_tag("table", "table", inspect_context, inspect_parameters)

    def get_finder_table_row(self, inspect_context:UIInspectContext=None, inspect_parameters:InspectParameters=None):
        """
        Finder for table rows (including header row if existing)
        @param inspect_context Inspect context
        @param inspect_parameters Inspect parameters
        @return Finder
        """
        criteria = SeleniumCriteria(self)
        criteria.tag_name = "tr"
        
        res = BySeleniumFinder(criteria=criteria, description="table row")
        return self._build_result_finder(res, inspect_context, inspect_parameters)
    
    def get_finder_table_cell(self, inspect_context:UIInspectContext=None, inspect_parameters:InspectParameters=None):
        """
        Finder for table cells
        @param inspect_context Inspect context
        @param inspect_parameters Inspect parameters
        @return Finder
        """
        criteria = SeleniumCriteria(self)
        criteria.tag_name_pattern = "^th|td$"
        
        res = BySeleniumFinder(criteria=criteria, description="table cell")
        return self._build_result_finder(res, inspect_context, inspect_parameters)
    
    def get_finder_text(self, text, inspect_context:UIInspectContext=None, inspect_parameters:InspectParameters=None):
        if inspect_context is None:
            inspect_context = self.inspect_builder.context()
        
        res = ElseFinder(f"text '{text}'")
        
        res.set_next_finder(self.get_finder_by_node_text(text, inspect_context.without_ui_context().with_finder_type("text-node"), inspect_parameters))
        res.set_next_finder(self.get_finder_by_content_text(text, inspect_context.without_ui_context().with_finder_type("text-content"), inspect_parameters))
        
        return self._build_result_finder(res, inspect_context, inspect_parameters)
    
    def get_finder_text_containing(self, text, inspect_context:UIInspectContext=None, inspect_parameters:InspectParameters=None):
        if Tools.do_log(logger, logging.DEBUG):
            logger.debug(f"WARNING: the use of a finder for text containing '{text}' can be very slow, it should be limited to small zones")
        criteria = SeleniumCriteria(self)
        criteria.set_content_text_pattern(re.escape(text), False)
        
        res = BySeleniumFinder(criteria=criteria, description=f"text containing '{text}'")
        
        return self._build_result_finder(res, inspect_context, inspect_parameters)
    
    def get_finder_text_element(self, inspect_context:UIInspectContext=None, inspect_parameters:InspectParameters=None):
        return self._get_finder_text_element(None, inspect_context, inspect_parameters)
    
    def _get_finder_text_element(self, additional_criteria, inspect_context:UIInspectContext=None, inspect_parameters:InspectParameters=None):
        criteria_text = self._get_criteria_text_element()
        if additional_criteria is None:
            res = BySeleniumFinder(criteria=criteria_text, description="text element")
        else:
            criteria = AndCriteria()
            criteria.add_criteria(criteria_text)
            criteria.add_criteria(additional_criteria)
            res = BySeleniumFinder(criteria=criteria, description=f"text element verifying [{additional_criteria}]")
        
        return self._build_result_finder(res, inspect_context, inspect_parameters)
    
    def get_finder_symbol(self, expected_symbol, inspect_context:UIInspectContext=None, inspect_parameters:InspectParameters=None):
        # Build pattern for Title symbol
        pattern_title = r"^Title\('(.*)'\)$"
        
        matcher = re.match(pattern_title, expected_symbol)
        if matcher:
            interpreter = SessionContext.instance().text_interpreter
            verifier = SessionContext.instance().text_verifier
            
            # Get and interpret title
            title = matcher.group(1)
            title = interpreter.interpret(title)
            
            # Create appropriate finder whereas title contain a verify section or not
            if interpreter.contains_interpret(title):
                # Criteria to find an element with a title
                criteria = SeleniumCriteria(self)
                criteria.set_attribute_pattern("title", ".+")
                
                # Criteria to then check title
                criteria_check = SeleniumCriteriaAndVerifier(criteria, verifier)
                criteria_check.set_attribute("title", title)
                
                # Return finder
                return BySeleniumFinder(criteria=criteria_check, description=f"titled as '{title}'")
            else:
                criteria = SeleniumCriteria(self)
                criteria.set_attribute("title", title)
                return BySeleniumFinder(criteria=criteria, description=f"titled with '{title}'")
        else:
            raise TechnicalException(f"Unmanaged symbol '{expected_symbol}'")
    
    def get_finder_zone(self, zone_name, inspect_context:UIInspectContext=None, inspect_parameters:InspectParameters=None):
        res = ThenFinder(f"zone '{zone_name}'")
        
        # Find label
        res.set_next_finder( self.get_finder_text(zone_name, inspect_context.without_ui_context(), inspect_parameters) )
        
        # Find div in parents
        criteria = SeleniumCriteria(self)
        criteria.tag_name = "div"
        res.set_next_finder(FindType.InParents, BySeleniumFinder(criteria=criteria, description="div parent") )
        
        return self._build_result_finder(res, inspect_context, inspect_parameters)
    
    
    
    
    


# -*- coding: utf-8 -*-

#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################


from holado_test.behave.behave import *  # @UnusedWildImport
from holado_value.common.tables.value_table_manager import ValueTableManager
from holado_test.scenario.step_tools import StepTools
from holado_binary.ipc.bit_series import BitSeries
import copy
from holado_binary.ipc import bit_series
from holado_value.common.tables.comparators.table_2_value_table_with_header_comparator import Table2ValueTable_WithHeaderComparator
import logging
from holado_core.common.tables.comparators.table_with_header_comparator import TableWithHeaderComparator
from holado_test.behave.scenario.behave_step_tools import BehaveStepTools
from holado.common.context.session_context import SessionContext
from holado_core.common.exceptions.functional_exception import FunctionalException
from holado_value.common.tables.converters.value_table_converter import ValueTableConverter

logger = logging.getLogger(__name__)



def __get_scenario_context():
    return SessionContext.instance().get_scenario_context()

def __get_text_interpreter():
    return __get_scenario_context().get_text_interpreter()

def __get_variable_manager():
    return __get_scenario_context().get_variable_manager()


def __verify_table_is_bit_series_table(table):
    if not ValueTableManager.verify_table_is_x_table(table, "Name", "Bit length", "Type", "Value", raise_exception=False) \
            and not ValueTableManager.verify_table_is_x_table(table, "Name", "Bit length", "Type", raise_exception=False):
        raise FunctionalException("Table header has to be [Name, Bit length, Type, Value] or [Name, Bit length, Type]")


@Step(r"(?P<var_name>{Variable}) = bit series")
def step_impl(context, var_name):
    var_name = StepTools.evaluate_variable_name(var_name)
    table = BehaveStepTools.convert_step_table_2_value_table_with_header(context.table)
    
    __verify_table_is_bit_series_table(table)
    bs = BitSeries(bit_sections_list = ValueTableConverter.convert_table_2_list_of_tuples(table, as_generator=True))
    
    __get_variable_manager().register_variable(var_name, bs)

@Step(r"fill bit series (?P<bs_var_name>{Variable}) from hexadecimal string (?P<hex_str>{Str})(?: \((?P<padding_side>right|left) padded\))?")
def step_impl(context, bs_var_name, hex_str, padding_side):
    bs = StepTools.evaluate_variable_value(bs_var_name)
    hex_str = StepTools.evaluate_scenario_parameter(hex_str)
    right_padded = padding_side == "right"
    
    bs.from_hex(hex_str, right_padded=right_padded)

@Step(r"(?P<var_name>{Variable}) = convert bit series (?P<bs_var_name>{Variable}) to hexadecimal string(?: \((?P<padding_side>right|left) padded\))?")
def step_impl(context, var_name, bs_var_name, padding_side):
    var_name = StepTools.evaluate_variable_name(var_name)
    bs = StepTools.evaluate_variable_value(bs_var_name)
    right_padding = padding_side == "right"
    
    res = bs.to_hex(right_padding=right_padding)
    
    __get_variable_manager().register_variable(var_name, res)

@Step(r"(?P<var_name>{Variable}) = convert bit series (?P<bs_var_name>{Variable}) to name/value table")
def step_impl(context, var_name, bs_var_name):
    var_name = StepTools.evaluate_variable_name(var_name)
    bs = StepTools.evaluate_variable_value(bs_var_name)
    
    res = bit_series.convert_bit_series_to_name_value_table(bs)
    
    __get_variable_manager().register_variable(var_name, res)

@Then(r"hexadecimal string (?P<hex_str>{Str}) is bit series")
def step_impl(context, hex_str):
    hex_str = StepTools.evaluate_scenario_parameter(hex_str)
    expected_table = BehaveStepTools.convert_step_table_2_value_table_with_header(context.table)
    
    __verify_table_is_bit_series_table(expected_table)
    
    # Build a table representing the bit series in hexadecimal string
    declare_table = copy.copy(expected_table)
    declare_table.remove_column(name="Value")
    bs = BitSeries(bit_sections_list=ValueTableConverter.convert_table_2_list_of_tuples(declare_table, as_generator=True))
    bs.from_hex(hex_str)
    obtained_table = bit_series.convert_bit_series_to_table(bs)
    
    comparator = Table2ValueTable_WithHeaderComparator()
    comparator.equals(obtained_table, expected_table)

@Then(r"hexadecimal string (?P<hex_str>{Str}) is bit series (?P<var_bit_series>{Variable})")
def step_impl(context, hex_str, var_bit_series):
    hex_str = StepTools.evaluate_scenario_parameter(hex_str)
    expected_bs = StepTools.evaluate_variable_value(var_bit_series)
    
    # Build a table representing the bit series in hexadecimal string
    bs = copy.copy(expected_bs)
    bs.from_hex(hex_str)
    obtained_table = bit_series.convert_bit_series_to_table(bs)
    
    expected_table = bit_series.convert_bit_series_to_table(expected_bs)

    comparator = TableWithHeaderComparator()
    comparator.equals(obtained_table, expected_table)

@Then(r"bit series (?P<var_bit_series>{Variable}) is")
def step_impl(context, var_bit_series):
    bs = StepTools.evaluate_variable_value(var_bit_series)
    expected_table = BehaveStepTools.convert_step_table_2_value_table_with_header(context.table)
    
    __verify_table_is_bit_series_table(expected_table)
    
    obtained_table = bit_series.convert_bit_series_to_table(bs)
    
    comparator = Table2ValueTable_WithHeaderComparator()
    comparator.equals(obtained_table, expected_table)






#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
#
# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

import logging
from holado_ui_selenium.ui.gui.selenium.inspectors.selenium_inspector import SeleniumInspector
from holado_ui.ui.inspectors.tools.ui_inspect_context import UIInspectContext
from holado_core.common.inspectors.tools.inspect_parameters import InspectParameters
from holado_ui_selenium.ui.gui.selenium.criterias.selenium_criteria import SeleniumCriteria
from holado_ui_selenium.ui.gui.selenium.finders.by_selenium_finder import BySeleniumFinder

logger = logging.getLogger(__name__)



class AngularSeleniumInspector(SeleniumInspector):
    """ Selenium inspector with Angular support.
    """
    
    def __init__(self):
        super().__init__("angular")
    
    @SeleniumInspector.default_inspect_builder.getter  # @UndefinedVariable
    def default_inspect_builder(self):
        res = super().default_inspect_builder
    
        res.default_parameters.add_finder_types([
            "mat-select"
            ])
        
        return res
    
    def get_finder_select(self, inspect_context:UIInspectContext=None, inspect_parameters:InspectParameters=None):
        if inspect_context is None:
            inspect_context = self.inspect_builder.context()
        
        return self.get_finder_by_tag("mat-select", inspect_context.with_finder_type("mat-select"), inspect_parameters)
    
    def get_finder_select_option(self, value, inspect_context:UIInspectContext=None, inspect_parameters:InspectParameters=None):
        """
        @param value Option value
        @param iContext Inspect context
        @param iParameters Inspect parameters
        @return Finder for the select option with given value
        """
        criteria = SeleniumCriteria(self)
        criteria.tag_name = "mat-option"
        criteria.set_attribute("role", "option")
        criteria.content_text = value
        
        res = BySeleniumFinder(criteria, f"select option '{value}'")
        return self._build_result_finder(res, inspect_context, inspect_parameters)
    
    
    
    
    


#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
#
# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

from holado_core.common.criterias.tools.criteria_context import CriteriaContext
from holado_core.common.criterias.tools.criteria_parameters import CriteriaParameters
import logging

logger = logging.getLogger(__name__)



class OrCriteria(object):
    """ Criteria that makes a logical or between added criterias.
    """
    
    def __init__(self):
        self.__criterias = []
    
    @property
    def criterias(self):
        return self.__criterias
    
    def add_criteria(self, criteria):
        """
        Add a criteria in or validation.
        @param criteria Criteria
        """
        self.__criterias.append(criteria)
        
    def validate_element(self, element, criteria_context: CriteriaContext, criteria_parameters: CriteriaParameters):
        return super().validate_element(element, criteria_context, criteria_parameters) \
            and self.__validate_criterias(element, criteria_context, criteria_parameters)

    def __validate_criterias(self, element, criteria_context, criteria_parameters):
        if len(self.criterias) == 0:
            return True
        
        res = False
        sub_parameters = criteria_parameters.get_next_level()
        
        for criteria in self.criterias:
            if criteria.validate(element, criteria_context, sub_parameters):
                res = True
                break
        
        return res

    def __str__(self):
        res_criterias = []
        for index, criteria in self.criterias:
            res_criterias.append(f"{index}:{{{criteria}}}")
        return super().__str__() + "&{" + "|".join(res_criterias) + "}"
    
    
    



#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

import logging
from holado_db.tools.db.clients.sqlite.sqlite_client import SQLite3Client
from holado_db.tools.db.clients.postgresql.postgresql_client import PostgreSQLClient
from holado_core.common.exceptions.technical_exception import TechnicalException
from holado_core.common.handlers.abstracts.get_or_create import GetOrCreateVariableObject
from holado_core.common.tools.tools import Tools
from holado.common.context.session_context import SessionContext


logger = logging.getLogger(__name__)



class DBManager(GetOrCreateVariableObject):
    """
    Manage DB clients, agnostic to managed DB.
    
    It manages an instance by thread for DB types (ex: sqlite3) which clients don't support multithreading.
    Note: Even if sqlite3 can support multithreading (with 'connect' parameter 'check_same_thread=False'),
          current implementation of DBManager keeps sqlite3 default behavior.
    """
    
    def __init__(self, name):
        super().__init__(name if name else "DBManager")
        
        self.__func_query_manager = None
        self.__db_info_by_db_type = {}
    
    def initialize(self, func_variable_manager, func_query_manager):
        super().initialize(func_variable_manager)
        self.__func_query_manager = func_query_manager
        
        if SQLite3Client.is_available():
            self.register_db_client("sqlite3", SQLite3Client, do_support_multithreading=False)
        else:
            if Tools.do_log(logger, logging.DEBUG):
                logger.debug("SQLite3 package is not installed, this type of DB is not available")
        if PostgreSQLClient.is_available():
            self.register_db_client("postgresql", PostgreSQLClient)
        else:
            if Tools.do_log(logger, logging.DEBUG):
                logger.debug("PostgreSQL package is not installed, this type of DB is not available")
    
    @property
    def __query_manager(self):
        return self.__func_query_manager()
    
    def register_db_client(self, db_type, new_client_func, do_support_multithreading=True):
        self.__db_info_by_db_type[db_type] = (new_client_func, do_support_multithreading)
        if Tools.do_log(logger, logging.DEBUG):
            logger.debug(f"Registered DB type '{db_type}'")
    
    def get_or_create(self, name, *args, **kwargs):
        # Manager DB types that need a client by thread
        db_type = self.__find_db_type(args, kwargs)
        if db_type is not None and not self.__db_info_by_db_type[db_type][1]:
            thread_db_manager = SessionContext.instance().multitask_manager.get_thread_context().db_manager
            if thread_db_manager is not self:
                return thread_db_manager.get_or_create(name, *args, **kwargs)
            else:
                # Prefix name by thread UID, since same variable manager is shared by all thread context DBManagers
                thread_uid = SessionContext.instance().multitask_manager.get_thread_uid()
                real_name = f"{thread_uid}-{name}"
        else:
            real_name = name
        
        return super().get_or_create(real_name, *args, **kwargs)
    
    def __find_db_type(self, args, kwargs):
        if kwargs and 'db_type' in kwargs:
            return kwargs['db_type']
        elif args and len(args) > 0 and args[0] in self.__db_info_by_db_type:
            return args[0]
        else:
            return None
    
    def _goc_new_object(self, name, db_type, connect_kwargs):
        res = self.new_client(name, db_type, connect_kwargs)
        res.connect()
        return res
    
    def new_client(self, name, db_type, connect_kwargs):
        if db_type in self.__db_info_by_db_type:
            try:
                res = self.__db_info_by_db_type[db_type][0](name, connect_kwargs)
            except Exception as exc:
                raise TechnicalException(f"Failed to create client for DB type '{db_type}' with connect parameters: {connect_kwargs}") from exc
        else:
            raise TechnicalException(f"Unmanaged DB type '{db_type}'")
        
        res.query_builder = self.__query_manager.new_default_query_builder(db_type, res)
        return res
        
        
        
        



#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

from builtins import super
import abc
import logging
from holado_core.common.tools.tools import Tools
from holado_python.common.tools.datetime import DateTime

logger = logging.getLogger(__name__)


class BaseBlock(object):
    '''
    Base class for blocks
    '''
    __metaclass__ = abc.ABCMeta
    
    def __init__(self, name, ID=None):
        super().__init__()
        self.__id = ID if ID is not None else name
        self.__name = name
        
        self.__start_dt = None
        self.__end_dt = None

    @property
    def ID(self):
        return self.__id
    
    @ID.setter
    def ID(self, id_):
        self.__id = id_

    @property
    def name(self):
        return self.__name

    @property
    def start_datetime(self):
        return self.__start_dt

    @property
    def end_datetime(self):
        return self.__end_dt

    @property
    def duration(self):
        if self.__start_dt is not None and self.__end_dt is not None:
            return self.__end_dt - self.__start_dt
        else:
            return None
    
    @abc.abstractmethod
    def process(self):
        raise NotImplementedError

    def _process_start(self):
        self.__start_dt = DateTime.now()

    def _process_end(self):
        self.__end_dt = DateTime.now()
    


class BaseScope(BaseBlock):
    '''
    Base class for scope blocks
    '''
    __metaclass__ = abc.ABCMeta
    
    def __init__(self, name, ID=None):
        super().__init__(name, ID=ID)
        self.__content = []
    
    def process(self, *args, **kwargs):
        """
        Process the scope.
        Returns last scope step result
        """
        res = None
        if Tools.do_log(logger, logging.DEBUG):
            logger.debug(f"Processing scope [{self.name}]: begin")
        
        self._process_start()
        try:
            res = self._process_blocks(*args, **kwargs)
        finally:
            self._process_end()
            if Tools.do_log(logger, logging.DEBUG):
                logger.debug(f"Processing scope [{self.name}]: end")

        return res
    
    def _process_blocks(self, *args, **kwargs):
        """
        Process inner blocks of the scope.
        Returns last scope step result
        """
        res = None
        for block in self.__content:
            res = block.process(*args, **kwargs)
        return res
    
    def add_block(self, block):
        self.__content.append(block)
        
    def add_method_call(self, func, *args, **kwargs):
        from holado_core.common.block.block_method import BlockMethod
        self.add_block( BlockMethod(func, *args, **kwargs) )
        
    



#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of self software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and self permission notice shall be included in all copies or substantial portions of the Software.
#
# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

import logging
import weakref
from holado_core.common.exceptions.technical_exception import TechnicalException
import queue
from holado_core.common.finders.tools.find_parameters import FindParameters
from holado_core.common.exceptions.element_exception import NoSuchElementException,\
    TooManyElementsException
from holado_core.common.finders.tools.enums import FindType
from holado_ui.ui.gui.handlers.label_gui_context import LabelGUIContext
from holado_ui.ui.gui.drivers.gui_driver import GUIDriver
from holado_ui.ui.gui.drivers.gui_internal_api import GUIInternalAPI
from holado_scripting.text.verifier.text_verifier import TextVerifier
import abc
from holado_core.common.tools.tools import Tools

logger = logging.getLogger(__name__)



class GUIWindow(object):
    """ Base class for GUI windows.
    """
    __metaclass__ = abc.ABCMeta
    
    def __init__(self):
        self.__driver_weakref = weakref.ref(None)
        
        self.__inspect_builder = None
        self.__find_builder = None
        
        self.__inspector = None
        self.__actor = None

        self.__current_container = None
        self.__popup_stack = queue.LifoQueue()
    
    def initialize(self, driver:GUIDriver):
        """
        Initialize window with given driver
        @param driver UI driver
        """
        self.__driver_weakref = weakref.ref(driver)
        
        self.__inspector = self._initialize_inspector()
        self.__actor = self._initialize_actor()

        # Must be after inspector initialize, so that default inspect builder is accessible throw inspector
        self.__inspect_builder = self._initialize_inspect_builder()

        self.__current_container = self.get_window_container()
    
    @property
    def driver(self) -> GUIDriver:
        """
        @return driver
        """
        res = self.__driver_weakref()
        if res is None:
            raise TechnicalException("Driver has been cleared")
        return res
    
    @property
    def internal_api(self) -> GUIInternalAPI:
        """
        @return Internal API
        """
        return self.driver.internal_api
    
    @property
    def inspect_builder(self):
        """
        @return Inspector builder
        """
        return self.get_inspect_builder(False)
    
    def get_inspect_builder(self, from_inspector):
        """
        Used internally to prior inspect builder in window than in inspector
        @param from_inspector If request comes from inspector
        @return Inspector instance
        """
        if self.__inspect_builder is not None:
            return self.__inspect_builder
        elif not from_inspector:
            return self.inspector.get_inspect_builder(True)
        else:
            return None
    
    def _initialize_inspect_builder(self):
        """
        This method is used internally at window creation, in order to instantiate the appropriate inspect builder.
        By default return None, in order to use the default one initialized in Inspector. 
        @return a new inspect builder
        """
        return None
    
    @property
    def find_builder(self):
        """
        @return Find builder
        """
        return self.get_find_builder(False)
    
    def get_find_builder(self, from_inspector):
        """
        Used internally to prior find builder in window than in inspector
        @param from_inspector If request comes from inspector
        @return Inspector instance
        """
        if self.__find_builder is not None:
            return self.__find_builder
        elif not from_inspector:
            return self.inspector.get_find_builder(True)
        else:
            return None
    
    def _initialize_find_builder(self):
        """
        This method is used internally at window creation, in order to instantiate the appropriate find builder.
        By default return None, in order to use the default one initialized in Inspector. 
        @return a new find builder
        """
        return None
    
    @property
    def inspector(self):
        """
        @return Inspector instance
        """
        return self.__inspector
    
    def _initialize_inspector(self):
        """
        This method is used internally at window creation, in order to instantiate the appropriate inspector type.
        @return a new inspector instance
        """
        raise NotImplementedError

    @property
    def actor(self):
        """
        @return Actor instance
        """
        return self.__actor
    
    def _initialize_actor(self):
        """
        This method is used internally at window creation, in order to instantiate the appropriate actor type.
        @return a new actor instance
        """
        raise NotImplementedError

    def _get_text_verifier(self) -> TextVerifier:
        return self.driver.text_verifier
    
    @property
    def current_container(self):
        return self.__current_container
    
    @current_container.setter
    def current_container(self, current_container):
        """
        @param current_container the current container to set
        """
        if Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
            logger.trace(f"Set current container to [{current_container.get_complete_description()}]") 
        self.__current_container = current_container
    
    def update_current_container(self, raise_no_such_element=True):
        """
        Update current container, usually after a page refresh
        @param raise_no_such_element Raise if any expected element is not found
        """
        # If page behind popup (no popup in stack)
        if self.__popup_stack.empty():
            # Update current container
            self.current_container = self.get_window_container()
            if Tools.do_log(logger, logging.DEBUG):
                logger.debug("Updated current container with page container") 
        
        
        # If popup behind popup
        else:
            popup_name = self.__popup_stack.queue[-1]
            popup = self.find_popup(popup_name, raise_no_such_element)
            if popup is not None:
                self.current_container(self.get_popup_container(popup_name, popup))
                if Tools.do_log(logger, logging.DEBUG):
                    logger.debug(f"Updated current container with popup '{popup_name}' (popup stack: {self.__popup_stack.queue})")
            else:
                if Tools.do_log(logger, logging.DEBUG):
                    logger.debug(f"Unable to update current container  unable to find popup '{popup_name}' (popup stack: {self.__popup_stack.queue})")
                    
    def get_zone_holder(self, zone, zone_name):
        """
        Create a holder with zone element
        @param zone element corresponding to zone
        """
        return self.current_container.get_holder_for(zone, f"zone '{zone_name}'")
    
    def _enter_popup(self, popup_name, popup_holder):
        """
        Enter in popup
        """
        # Update current container
        self.current_container = popup_holder
        
        # Add popup in stack
        self.__popup_stack.put(popup_name)
        if Tools.do_log(logger, logging.DEBUG):
            logger.debug(f"Entered in popup '{popup_name}' -> Popup stack: {self.__popup_stack.queue}") 
    
    def _leave_popup(self, raise_no_such_element=True):
        """
        Go back to previous popup or page
        """
        # Remove closed popup from stack
        popup_name = self.__popup_stack.get()
        if Tools.do_log(logger, logging.DEBUG):
            logger.debug(f"Leaved popup '{popup_name}' -> Popup stack: {self.__popup_stack.queue}") 
        
        # Update current container
        self.update_current_container(raise_no_such_element)
    
    def is_any_popup_opened(self):
        """
        @return True if any popup is supposed to be opened
        """
        return not self.__popup_stack.empty()
    
    def wait_until_popup_is_closed(self, raise_no_such_element=True, timeout_seconds=None, popup_title=None):
        """
        Wait until popup is closed
        @param raise_no_such_element Raise if any expected element is not found
        @param timeout_seconds Timeout duration in seconds
        @param popup_title Popup title
        """
        # Define popup title
        if popup_title is None:
            popup_name = self.__popup_stack.queue[-1]
            if Tools.do_log(logger, logging.DEBUG):
                logger.debug(f"Waiting until popup '{popup_name}' is closed (popup stack: {self.__popup_stack.queue})...")
                 
            if popup_name != "{EMPTY}":
                popup_title = popup_name
            
        # Wait popup is closed
        self.actor.redo_while_not_null(
                self.actor.act_find_popup(popup_title),
                timeout_seconds)
        
        # Leave popup
        self._leave_popup(raise_no_such_element)
    
    def get_window_container(self):
        """
        @return Window container
        """
        raise NotImplementedError

    def _get_popup_container(self, popup_title, popup_element):
        raise NotImplementedError
    
    def click_on_text(self, ui_context, text, redo_while_no_such_element):
        """
        Click on given text
        @param ui_context Context
        @param text Text
        @param redo_while_no_such_element If True, retry while catching exception NoSuchElementException
        """
        parameters = FindParameters.default_with_raise().with_redo(redo_while_no_such_element).ignoring(NoSuchElementException)
        act = self.actor.act_find_and_click_on(self.inspector.get_finder_text(text, self.inspector.inspect_builder.context(ui_context), None),
                                               FindType.In)
        act.execute(self.current_container, parameters)
    
    def for_label_is_displayed(self, ui_context, label, text, raise_exception):
        """
        Check the value associated to a label
        @param ui_context Context
        @param label Label
        @param text Text associated to label
        @param raise_exception If True, raise an exception rather than returning False
        @return Check result
        """
        ui_context_label = LabelGUIContext.for_label(ui_context, label)
        act = self.actor.act_check_text_is_displayed(ui_context_label, text)
        return act.execute(self.current_container, FindParameters.default(raise_exception))
    
    def check_table_contains(self, ui_context, expected, raise_exception):
        """
        Check that the table contains a sub-table
        @param ui_context Context
        @param expected Value table
        @param raise_exception If True, raise an exception when verification is False
        @return Check result
        """
        raise NotImplementedError
    
    def check_table_displays(self, ui_context, expected, raise_exception):
        """
        Check that the table is equal to given value table
        @param ui_context Context
        @param expected Value table
        @param raise_exception If True, raise an exception when verification is False
        @return Check result
        """
        raise NotImplementedError
    
    def check_text_containing_x_is_displayed(self, ui_context, text):
        """
        Verify that a text containing given text is displayed.
        @param ui_context Context
        @param text Text to find
        """
        act = self.actor.act_check_text_containing_is_displayed(ui_context, text)
        act.execute(self.current_container, FindParameters.default_with_raise())
    
    def check_text_x_is_displayed(self, ui_context, text):
        """
        Verify that given text is displayed.
        @param ui_context Context
        @param text Text to find
        """
        act = self.actor.act_number_of_occurrences(self.inspector.get_finder_text(text, self.inspector.inspect_builder.context(ui_context), None), FindType.In)
        nb = act.execute(self.current_container)
        if nb < 1:
            raise NoSuchElementException(f"Unable to find text '{text}'")
        elif nb > 1:
            raise TooManyElementsException(f"text '{text}' was found {nb} times")
    
    def is_text_displayed(self, ui_context, text):
        """
        @param ui_context Context
        @param text Text to find
        @return True if text was found
        """
        act = self.actor.act_number_of_occurrences(self.inspector.get_finder_text(text, self.inspector.inspect_builder.context(ui_context), None), FindType.In)
        nb = act.execute(self.current_container)
        return (nb >= 1)
    
    def check_new_popup_is_opened(self, expected_popup_title=None, raise_exception=True):
        """
        @param expected_popup_title Title of the popup
        @param raise_exception If True, raise an exception rather than returning False
        @return True if a popup has appeared with expected title
        """
        raise NotImplementedError

    """
    @param expected_popup_title Title of the popup
    @param raise_no_such_exception If True, raise an exception rather than returning None
    @return Popup element
    """
    def find_popup(self, expected_popup_title, raise_no_such_exception):
        # Find popup
        res = self.inspector.get_finder_popup().find_in(self.get_window_container(), FindParameters.default(raise_no_such_exception))
        if not raise_no_such_exception and res is None:
            return None
        
        # Find and verify the title
        if expected_popup_title is not None:
            title = self.inspector.get_finder_popup_title().findIn(res, FindParameters.default_without_raise())
            if title != None:
                obtained_popup_title = self.internal_api.get_element_text(title)
            else:
                obtained_popup_title = ""
            
            
            # Check popup title
            if not self._get_text_verifier().check(expected_popup_title, obtained_popup_title):
                if raise_no_such_exception:
                    raise NoSuchElementException(f"Current popup does not have the right title (obtained: '{obtained_popup_title}'  expected: '{expected_popup_title}')")
                else:
                    res = None
        
        return res
    
    
    
    



#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################


#################################################
# GOAL: Tools when using a clone of HolAdo project.
#
# This file contains methods usefull to initialize environments using a clone of HolAdo project.
#
# USAGE: 
#   - Copy this file in projects using HolAdo.
#   - Define environment variable HOLADO_PATH with path to cloned HolAdo project.
#     If HOLADO_PATH is defined, sources of cloned HolAdo project are used, else installed holado package is used.
#################################################



import os
import sys


def insert_sys_path(path, index=0):
    """Insert a path in sys.path if it doesn't already exists.
    """
    if path not in sys.path:
        sys.path.insert(index, path)

def insert_sys_paths(list_paths, index=0):
    """Insert a list of path in sys.path if it doesn't already exists.
    """
    if list_paths:
        for path in list_paths:
            insert_sys_path(path, index=index)

def insert_holado_source_paths(with_test_behave=True):
    """Insert in sys.path all HolAdo source paths.
    If environment variable HOLADO_PATH is defined with path to HolAdo project, following paths are inserted in sys.path:
        - HOLADO_PATH/src: path to holado modules sources
        - HOLADO_PATH/tests/behave (if with_test_behave==True): path to holado test sources, needed by testing solutions
    """
    holado_path = os.getenv('HOLADO_PATH')
    if holado_path is None:
        try:
            import holado  # @UnusedImport
        except Exception as exc:
            if "No module named" in str(exc):
                raise Exception(f"If environment variable HOLADO_PATH is not defined with path to HolAdo project, 'holado' python package must be installed")
            else:
                raise exc
        else:
            # holado is installed, and all sources are already accessible
            pass
    else:
        print(f"Using HolAdo project installed in '{holado_path}'")
        # import traceback
        # print("".join(traceback.format_list(traceback.extract_stack())))
        # insert_sys_path(holado_path)
        insert_sys_path(os.path.join(holado_path, "src"))
        if with_test_behave:
            insert_sys_path(os.path.join(holado_path, "tests", "behave"))
    





#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

import logging
from holado_core.common.block.function import Function
from holado_core.common.exceptions.functional_exception import FunctionalException
from holado_core.common.tools.tools import Tools
from holado_multitask.multiprocessing.process import InterruptableProcess

logger = logging.getLogger(__name__)


class FunctionProcess(InterruptableProcess):
    '''
    Execute in a process given function with given arguments
    '''
    
    def __init__(self, target, args=None, kwargs=None, name = None, default_wait_timeout = None, register_process = True, delay_before_run_sec=None, delay_between_run_sec=None):
        super().__init__(name if name is not None else f"FunctionProcess({repr(target)})", default_wait_timeout = default_wait_timeout, register_process = register_process, delay_before_run_sec=delay_before_run_sec, delay_between_run_sec=delay_between_run_sec)
        
        if args is None:
            args = []
        if kwargs is None:
            kwargs = {}
        
        self._function = Function(target, *args, **kwargs)
        self.__error = None
        self.__result = None
        self.__interrupt_function = None
        self.__callback_function = None

    @property
    def error(self):
        return self.__error

    @property
    def result(self):
        return self.__result

    @property
    def interrupt_function(self):
        return self.__interrupt_function

    @interrupt_function.setter
    def interrupt_function(self, func):
        """Set the Function to call on interrupt"""
        self.__interrupt_function = func

    @property
    def callback_function(self):
        return self.__callback_function

    @callback_function.setter
    def callback_function(self, func):
        """Set the callback function"""
        self.__callback_function = func
 
    def run(self):
        # Start running by superclass run
        super().run()
        
        # self._raise_if_interrupted()
        self.wait_before_run()
        
        # self._raise_if_interrupted()
        try:
#             logging.debug("+++++++++ Launching function [{}({})]".format(repr(self._target), repr(self._args)))
            self.__result = self._function.run()
        except Exception as exc:
            logger.exception(f"[{self.name}] Exception catched during processed function [{self._function.represent()}]")
            self.__error = exc
        
        # self._raise_if_interrupted()
        try:
            if self.__callback_function:
                self.__callback_function.run(self.result, self.error)
        except Exception as exc:
            msg = f"[{self.name}] Exception catched while calling callback function:\n    callback_function: {repr(self.__callback_function)}\n    result: [{Tools.truncate_text(repr(self.result))}]\n    error: [{repr(self.error)}]"
            logger.exception(msg)
            raise FunctionalException(msg) from exc
        
        logger.info(f"[{self.name}] Process has finished")

    def throw_if_error(self):
        if self.error:
            raise FunctionalException(f"[{self.name}] Error during call of function [{repr(self._target)}({repr(self._args)})]") from self.error
    
    def interrupt(self):
        if not self.is_interrupted and self.is_alive():
            super().interrupt()
            if self.interrupt_function is not None:
                self.interrupt_function.run()



# -*- coding: utf-8 -*-

#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

import logging
import datetime
import dateutil.parser
import re
from holado_core.common.exceptions.functional_exception import FunctionalException
from holado_core.common.exceptions.technical_exception import TechnicalException
import os
from holado_core.common.tools.converters.converter import Converter
from enum import Enum
from holado_python.common.enums import ArithmeticOperator
from datetime import timedelta
from dateutil.relativedelta import relativedelta
import math
from decimal import Decimal
from holado_python.standard_library.typing import Typing

logger = logging.getLogger(__name__)


EPOCH_1970 = datetime.datetime(year=1970, month=1, day=1, tzinfo=datetime.timezone.utc)
EPOCH_JULIAN_CNES = datetime.datetime(year=1950, month=1, day=1, tzinfo=datetime.timezone.utc)
EPOCH_JULIAN_NASA = datetime.datetime(year=1968, month=5, day=24, tzinfo=datetime.timezone.utc)

FORMAT_DATETIME_ISO = '%Y-%m-%dT%H:%M:%S.%fZ'
FORMAT_DATETIME_ISO_SECONDS = '%Y-%m-%dT%H:%M:%SZ'
FORMAT_DATETIME_HUMAN_SECOND = '%Y-%m-%d %H:%M:%S'
FORMAT_TIME_ISO = '%H:%M:%S.%f'
FORMAT_TIME_ISO_SECONDS = '%H:%M:%S'

TIMEZONE_LOCAL = datetime.datetime.now().astimezone().tzinfo
TIMEZONE_UTC = datetime.timezone.utc
default_timezone = TIMEZONE_UTC


class DurationUnit(str, Enum):
    # Nanosecond = "nanosecond"
    Microsecond = "microsecond"
    Millisecond = "millisecond"
    Second = "second"
    Minute = "minute"
    Hour = "hour"
    Day = "day"
    Week = "week"
    Month = "month"
    Year = "year"
    


class DateTime(object):
    """ Tools to manipulate datetimes.
    Note: Naive and aware datetimes cannot be compared. Thus HolAdo methods always set datetime timezone when it is known.
          In methods unable to define the timezone (ex: str_2_datetime), it is highly recommended to specify 'tz' argument.
    """
    
    __is_system_time_in_tai = None
    __is_system_time_expected_in_tai = None
    __utc_to_tai_timedelta = None
    
    
    def __init__(self):
        pass
    
    @classmethod
    def is_system_time_in_tai(cls):
        """Return if system time is in TAI (International Atomic Time)"""
        if cls.__is_system_time_in_tai is None:
            cls.__is_system_time_in_tai = Converter.to_boolean(os.getenv("HOLADO_SYSTEM_TIME_IS_IN_TAI", False))
        return cls.__is_system_time_in_tai
    
    @classmethod
    def now(cls, tz=default_timezone):  # @UndefinedVariable
        return datetime.datetime.now(tz=tz)
    
    @classmethod
    def utcnow(cls):
        if cls.is_system_time_in_tai():
            return cls.now(TIMEZONE_UTC) - cls.utc_to_tai_timedelta()
        else:
            return cls.now(TIMEZONE_UTC)
    
    @classmethod
    def tainow(cls):
        if cls.is_system_time_in_tai():
            return cls.now(TIMEZONE_UTC)
        else:
            return cls.now(TIMEZONE_UTC) + cls.utc_to_tai_timedelta()
    
    @classmethod
    def utc_to_tai_timedelta(cls):
        """Return the timedelta to add to an UTC datetime to obtain a TAI datetime.
        Note: Currently, only years 2020-2029 are managed"""
        if cls.__utc_to_tai_timedelta is None:
            cls.__utc_to_tai_timedelta = datetime.timedelta(seconds=37)
        return cls.__utc_to_tai_timedelta
    
    @classmethod
    def apply_delta_on_datetime(cls, date:datetime, operator:ArithmeticOperator, duration, duration_unit:DurationUnit) -> datetime:
        args={duration_unit.value+'s':duration}
        if duration_unit in [DurationUnit.Millisecond]:
            delta = timedelta(**args)
        else:
            delta = relativedelta(**args)
        
        if operator == ArithmeticOperator.Subtraction: 
            res = date - delta
        elif operator == ArithmeticOperator.Addition:
            res = date + delta
        else:
            raise TechnicalException(f"Unmanaged operator '{operator.value}'")
        
        return res
    
    ### Conversions to/from str
    
    @classmethod
    def is_str_datetime(cls, dt_str, dt_format=None):
        """Return if a string is a datetime of given format"""
        res = True
        try:
            DateTime.str_2_datetime(dt_str, dt_format)
        except (dateutil.parser._parser.ParserError, ValueError):
            res = False
        return res

    @classmethod
    def str_2_datetime(cls, dt_str, dt_format=None, tz=None):
        """Convert string to datetime instance
        @param dt_format If defined, use this format to parse string rather than using first matching format
        @param tz If defined, return a datetime with this timezone. 
                  If string already defines a timezone, the datetime is converted to this timezone, 
                  else the string is supposed to represent a datetime in this timezone and timezone is set in result datetime.
        Note: Naive and aware datetimes cannot be compared. Thus HolAdo methods always set datetime timezone when it is known.
              It is highly recommended to specify 'tz' argument
        """
        res = None
        if dt_format is None:
            res = dateutil.parser.parse(dt_str)
        else:
            res = datetime.datetime.strptime(dt_str, dt_format)
        
        # Set timezone if needed
        if tz is not None:
            if res.tzinfo is None:
                res = res.replace(tzinfo=tz)
            else:
                res = res.astimezone(tz)
        
        return res

    @classmethod
    def str_2_timedelta(cls, dur_str):
        """Convert string to timedelta instance"""
        res = None
        dur_regex = re.compile(r"(\d+(?:\.\d*)?) (?:s|sec|seconds)")
        r = dur_regex.match(dur_str)
        if r:
            res = datetime.timedelta(seconds=float(r.group(1)))
        else:
            raise FunctionalException(r"Duration string must match pattern '\d+(\.\d*)? (s|sec|seconds)'")
        return res
    
    @classmethod
    def datetime_2_str(cls, dt_value, dt_format=None):
        res = None
        if dt_format is None:
            res = datetime.datetime.isoformat(dt_value)
        else:
            res = dt_value.strftime(dt_format)
        return res
    
    @classmethod
    def _get_datetime(cls, dt, tz_in_str=None):
        res = None
        if dt is not None:
            if isinstance(dt, datetime.datetime):
                res = dt
            elif isinstance(dt, str):
                res = DateTime.str_2_datetime(dt, tz=tz_in_str)
            else:
                raise TechnicalException(f"Datetime can be a datetime, a str containing a datetime, or None")
        return res
    
    @classmethod
    def _get_epoch_datetime(cls, epoch, tz_in_str=None):
        res = None
        if epoch is not None:
            if isinstance(epoch, datetime.datetime):
                res = epoch
            elif isinstance(epoch, str):
                if epoch in ['EPOCH_1970', 'EPOCH_JULIAN_CNES', 'EPOCH_JULIAN_NASA']:
                    res = eval(epoch)
                else:
                    res = DateTime.str_2_datetime(epoch, tz=tz_in_str)
            else:
                raise TechnicalException(f"Epoch can be a datetime, a str containing a datetime, a str containing an EPOCH name, or None")
        return res
        
    @classmethod
    def _get_timestamp(cls, ts):
        res = None
        if ts is not None:
            if isinstance(ts, float) or isinstance(ts, int):
                res = ts
            elif isinstance(ts, str) and (Converter.is_float(ts) or Converter.is_integer(ts)):
                res = Converter.to_float(ts)
            else:
                raise TechnicalException(f"Timestamp must be a float|int or a string containing a float|int (obtained type: {Typing.get_object_class_fullname(ts)})")
        return res
    
    
    ### Conversions to/from timestamp
    
    @classmethod
    def datetime_to_timestamp(cls, dt, epoch=None):
        dt = cls._get_datetime(dt)
        epoch = cls._get_epoch_datetime(epoch)
        
        if epoch is None:
            return dt.timestamp()
        else:
            return (dt - epoch).total_seconds()
    
    @classmethod
    def timestamp_to_datetime(cls, ts, epoch=None):
        """
        Convert timestamp to UTC datetime
        """
        ts = cls._get_timestamp(ts)
        epoch = cls._get_epoch_datetime(epoch)
        
        res = datetime.datetime.fromtimestamp(ts, tz=datetime.timezone.utc)
        if epoch is not None:
            res = epoch + (res - EPOCH_1970)
        return res
    
    
    ### Conversions to/from (seconds, nanos)
    
    @classmethod
    def datetime_to_seconds_nanos(cls, dt, epoch=None):
        ts = cls.datetime_to_timestamp(dt, epoch)
        return cls.timestamp_to_seconds_nanos(ts)
    
    @classmethod
    def seconds_nanos_to_datetime(cls, seconds, nanos, epoch=None, round_func=round):
        """
        Convert timestamp to UTC datetime
        @param round_func: defines the rounding method applied to microseconds (default: round ; possible methods: round, math.ceil, math.trunc)
        """
        ts = cls.seconds_nanos_to_timestamp(seconds, nanos, round_func=round_func)
        return cls.timestamp_to_datetime(ts, epoch)
    
    @classmethod
    def timestamp_to_seconds_nanos(cls, ts):
        return round(ts), round(ts * 1e6 % 1e6)
    
    @classmethod
    def seconds_nanos_to_timestamp(cls, seconds, nanos, cast_func=float, round_func=round):
        """
        Convert (seconds, nanos) to timestamp.
        By default, it returns a timestamp as float with nanos rounded to microsecond precision.
        To keep nanosecond precision, set cast_func=None and round_func=None, it will returns a timestamp as decimal.Decimal.
        @param cast_func: Type of the returned value (default: float). If None, no cast is applied, and result is of type decimal.Decimal, usefull to keep nanosecond precision.
        @param round_func: defines the rounding method applied to microseconds (default: round ; possible methods: round, math.ceil, math.trunc). If None, value is not rounded.
        """
        res_dec = Decimal(seconds) + Decimal(nanos) / Decimal(1e9)
        if round_func is not None:
            res_dec = round_func(res_dec * 1000000) / 1000000
        if cast_func is not None:
            return cast_func(res_dec)
        else:
            return res_dec
    
    
    ### Conversions to/from Julian date
    
    @classmethod
    def datetime_to_julian_date(cls, dt, epoch=None):
        if epoch is None:
            return cls.datetime_to_julian_date(dt, epoch=EPOCH_JULIAN_CNES) + 2433282.5
        else:
            td = (dt - epoch)
            return td.days + td.seconds / (3600 * 24)
        
    @classmethod
    def julian_date_to_datetime(cls, jd, epoch=None, precision_nb_digits=6):
        """ Convert a Julian date to a datetime instance
        @param precision_nb_digits Precision of the Julian date (default: 6 = microseconds)
        """
        if epoch is None:
            return cls.julian_date_to_datetime(jd - 2433282.5, epoch=EPOCH_JULIAN_CNES)
        else:
            jd_decimal, jd_days = math.modf(jd)
            jd_seconds = round(jd_decimal * 3600 * 24, precision_nb_digits)
            td = datetime.timedelta(days=round(jd_days), seconds=jd_seconds)
            return epoch + td
        
    @classmethod
    def datetime_utc_to_tai(cls, dt, dt_format=None):
        """ Convert a datetime in UTC to a datetime in TAI
        @param dt Datetime in UTC
        @param dt_format Format to use if 'dt' is a string (can be None for format autodetection)
        """
        dt_src = dt
        if isinstance(dt_src, str):
            dt_src = cls.str_2_datetime(dt_src, dt_format, tz=TIMEZONE_UTC)
        
        return dt_src + DateTime.utc_to_tai_timedelta()
    
    @classmethod
    def datetime_tai_to_utc(cls, dt, dt_format=None):
        """ Convert a datetime in TAI to a datetime in UTC
        @param dt Datetime in TAI
        @param dt_format Format to use if 'dt' is a string (can be None for format autodetection)
        """
        dt_src = dt
        if isinstance(dt_src, str):
            dt_src = cls.str_2_datetime(dt_src, dt_format, tz=TIMEZONE_UTC)
        
        return dt_src - DateTime.utc_to_tai_timedelta()
        
    @classmethod
    def truncate_datetime(cls, dt, precision_nanoseconds=1):
        total_us = cls.datetime_to_timestamp(dt) * 1e9
        trunc_total_us = total_us // precision_nanoseconds * precision_nanoseconds
        return cls.timestamp_to_datetime(trunc_total_us / 1e9)
        



#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2023 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

import logging
from holado_db.tools.db.clients.base.db_client import DBClient
from holado_core.common.tables.table_with_header import TableWithHeader
from holado_core.common.tools.tools import Tools
from holado_db.tools.db.clients.postgresql.postgresql_audit import PostgreSQLTriggerAuditManager
from holado_db.tools.db.query.pypika.pypika_query_builder import PypikaQueryBuilder
from holado_core.common.exceptions.technical_exception import TechnicalException

logger = logging.getLogger(__name__)

try:
    import psycopg
    with_psycopg = True
except Exception as exc:
    if Tools.do_log(logger, logging.DEBUG):
        logger.debug(f"PostgreSQLClient is not available. Initialization failed on error: {exc}")
    with_psycopg = False

if PypikaQueryBuilder.is_available():
    from pypika.terms import Function, LiteralValue
    
    # from pypika import Function
    class JsonbSet(Function):
        def __init__(self, name, *args, **kwargs):
            super().__init__('jsonb_set', LiteralValue(name), *args, **kwargs)


class PostgreSQLClient(DBClient):
    @classmethod
    def is_available(cls):
        return with_psycopg

    def __init__(self, name, connect_kwargs):
        super().__init__(name if name else 'PostgreSQL', connect_kwargs)
        
    def _new_audit_manager(self):
        return PostgreSQLTriggerAuditManager("Audit with triggers", self)
        
    def _get_base_exception_type(self):
        return psycopg.Error
    
    def _connect(self, **kwargs):
        return psycopg.connect(**kwargs)
        
    def _get_sql_placeholder(self):
        return "%s"
        
    def exist_table(self, table_name):
        result = self.execute(f"SELECT tablename FROM pg_tables WHERE schemaname = 'public' AND tablename = '{table_name}'")
        return result is not None
        
    def set_or_update_json_key_value(self, table_name, field_name, json_key, json_value, where_data: dict=None, where_compare_data: list=None):
        # Help on JSON column: https://www.databasestar.com/postgresql-json/#How_to_Update_JSON_Data_in_PostgreSQL
        result = self.select(table_name, where_data=where_data, where_compare_data=where_compare_data, sql_return=field_name)
        is_set = isinstance(result, TableWithHeader) and result[0][0].content is not None
        if not is_set:
            self.update(table_name, {field_name: f'{{"{json_key}":"{json_value}"}}'}, where_data=where_data, where_compare_data=where_compare_data, do_commit=True)
        else:
            result = self.select(table_name, where_data=where_data, where_compare_data=where_compare_data, sql_return=f"{field_name}")
            is_key_set = False
            if isinstance(result, TableWithHeader) and result[0][0].content is not None:
                is_key_set = json_key in result[0][0].content
            if is_key_set:
                if not PypikaQueryBuilder.is_available():
                    # Note: This case cannot appear currently (04/08/2025), but can appear if another QueryBuilder is managed that is not using pypika
                    raise TechnicalException(f"Missing dependence: pypika")
                self.update(table_name, {field_name: JsonbSet(f'{field_name}', f'{{"{json_key}"}}', f'"{json_value}"')}, where_data=where_data, where_compare_data=where_compare_data, do_commit=True)
            else:
                self.update(table_name, {field_name: f'{field_name} || {{"{json_key}":"{json_value}"}}'}, where_data=where_data, where_compare_data=where_compare_data, do_commit=True)
    
    




#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

import logging
from holado_scripting.text.interpreter.functions.function_dynamic_value import FunctionDynamicValue
import re
from holado_scripting.text.interpreter.exceptions.interpreter_exception import NotExistingFunctionTextInterpreterException
from holado_scripting.text.interpreter.functions.function_cast import FunctionCast
from holado_core.common.exceptions.technical_exception import TechnicalException
from holado_scripting.text.interpreter.functions.function_to_base_64 import FunctionToBase64
from holado_scripting.text.interpreter.functions.function_to_hex import FunctionToHex
from holado_scripting.text.interpreter.functions.function_hex_to_int import FunctionHexToInt
from holado_scripting.text.interpreter.functions.function_to_bytes import FunctionToBytes
from holado_scripting.text.interpreter.functions.function_hex_to_bytes import FunctionHexToBytes
from holado_core.common.tools.tools import Tools
from holado_scripting.text.base.text_inspecter import TextInspecter
from holado_scripting.text.interpreter.functions.function_escape_all_bytes import FunctionEscapeAllBytes
from holado_scripting.common.tools.evaluate_parameters import EvaluateParameters
from holado_scripting.text.interpreter.functions.function_exists_variable import FunctionExistsVariable
from holado_scripting.text.interpreter.functions.function_convert import FunctionConvert
from holado_python.common.tools.datetime import DateTime
from holado_python.standard_library.typing import Typing
from holado_scripting.text.interpreter.functions.function_to_string import FunctionToString
from holado_scripting.text.interpreter.functions.function_apply_function import FunctionApplyFunction

logger = logging.getLogger(__name__)


class TextInterpreter(TextInspecter):

    def __init__(self):
        super().__init__()
        
        self.__text_verifier = None
        
        # self.__regex_to_be_interpreted = re.compile(r"^\$\{((?:(?!\$\{)[^\}])*)\}$")
        self.__regex_to_be_interpreted = re.compile(r"^\$\{(.*)\}$")
        
    def initialize(self, var_manager, expression_evaluator, text_verifier, dynamic_text_manager):
        super().initialize(var_manager, expression_evaluator)
        self.__text_verifier = text_verifier
        
        self.__register_default_functions(dynamic_text_manager)

    def interpret(self, str_to_interpret, eval_params=EvaluateParameters.default_without_raise(), log_level=logging.DEBUG):
        """
        @summary: Interpret given text by replacing all occurrences of "${...}" by corresponding values.
        @param str_to_interpret: String - Text to interpret
        @return: Interpreted result
        """
        
        if str_to_interpret is None:
            return None
        
        if Tools.do_log(logger, logging.TRACE, max_log_level=log_level):  # @UndefinedVariable
            logger.log(Tools.do_log_level(logging.TRACE, log_level), "Interpreting '{}'...".format(str_to_interpret))  # @UndefinedVariable
        
        # Interpret and get result as object
        res, _ = self.__interpret(str_to_interpret, eval_params=eval_params, log_level=log_level)

        if Tools.do_log_if_objects_are_different(logger, log_level, str_to_interpret, res):
            logger.log(log_level, "Interpreting '{}' -> [{}] (type: {})".format(str_to_interpret, res, type(res)))
            
        return res

    def _interpret_sections(self, str_to_interpret, eval_params=EvaluateParameters.default_without_raise(), log_level=logging.DEBUG):
        """
        @summary: Interpret given text by replacing all occurrences of "${...}" by variables containing corresponding values.
        @param str_to_interpret: String - Text to interpret
        @return: Interpreted result
        """
        
        if str_to_interpret is None:
            return None
        
        if Tools.do_log(logger, logging.TRACE, max_log_level=log_level):  # @UndefinedVariable
            logger.log(Tools.do_log_level(logging.TRACE, log_level), "Interpreting sections in '{}'...".format(str_to_interpret))  # @UndefinedVariable
        
        # Interpret and get result as object
        res, locals_ = self.__interpret(str_to_interpret, replace_sections_by_variables=True, eval_params=eval_params, log_level=log_level)

        if Tools.do_log_if_objects_are_different(logger, log_level, str_to_interpret, res):
            logger.log(log_level, f"Interpreting sections in '{str_to_interpret}' -> [{res}] (type: {type(res)}, locals: {locals_})")
        return res, locals_

    def __interpret(self, str_to_interpret, replace_sections_by_variables=False, eval_params=EvaluateParameters.default_without_raise(), log_level=logging.DEBUG):
        res = None
        locals_ = {}
        
        # If the string is to be interpreted
        section_indexes = self._search_all_interpret_sections(str_to_interpret)
        if (0,len(str_to_interpret)) in section_indexes:
            r_match = self.__regex_to_be_interpreted.match(str_to_interpret)
            if r_match:
                str_under_interpret = r_match.group(1).strip()
                res = self.__interpret_text(str_under_interpret, eval_params=eval_params, log_level=log_level)
                # Text is not interpretable ; expect it will be the case later
                if res == str_under_interpret:
                    res = str_to_interpret
                elif isinstance(res, str) and self.__text_verifier._has_function(res, with_interpreted=False):
                    res = f"${{{res}}}"
            else:
                raise TechnicalException(f"Failed to find interpret section whereas its indexes were found")
        else:
            res, locals_ = self.__find_and_interpret_sections(str_to_interpret, replace_sections_by_variables=replace_sections_by_variables, eval_params=eval_params, log_level=log_level)
            
        return res, locals_
    
    def __interpret_text(self, text, eval_params=EvaluateParameters.default_without_raise(), log_level=logging.DEBUG):  # @UndefinedVariable
        # Manage functions
        if self._has_function(text, with_interpreted=False):
            return self.__evaluate_function(text, eval_params=eval_params, log_level=log_level)
        
        # Manage evaluation
        _, res = self._expression_evaluator.evaluate_expression(text, eval_params=eval_params, log_level=Tools.do_log_level(log_level, logging.TRACE))  # @UndefinedVariable
        return res
    
    def __find_and_interpret_sections(self, str_to_interpret, replace_sections_by_variables=False, eval_params=EvaluateParameters.default_without_raise(), log_level=logging.DEBUG):
        res = str_to_interpret
        locals_ = {}
        
        offset = 0
        excluded_indexes = []
        while True:
            indexes = self._search_interpret_section(res, offset, excluded_indexes)
            if indexes is None:
                # Try again from beginning (in case of nested sections)
                indexes = self._search_interpret_section(res, 0, excluded_indexes)
                if indexes is not None:
                    offset = 0
                else:
                    # No more section to replace
                    break
                                
            ind_beg, ind_end = indexes
            sub_str = res[ind_beg:ind_end]
            
            # Interpret sub-section
            interpreted_value = self.interpret(sub_str, eval_params=eval_params, log_level=log_level)
            
            # Manage sections that aren't interpretable
            interpreted_str = f"{interpreted_value}"
            if interpreted_str == sub_str:
                excluded_indexes.append((ind_beg, ind_end))
                offset = ind_end
                continue
            
            # Define section replacement
            if replace_sections_by_variables and type(interpreted_value) not in [str, int, float, bytes]:
                var_name = self._variable_manager.new_local_variable_name()
                locals_[var_name] = interpreted_value
                section_replace_str = var_name
            else:
                section_replace_str = interpreted_str
            
            # Replace by interpreted value
            res = f"{res[0:ind_beg]}{section_replace_str}{res[ind_end:]}"
            
            # Prepare next replacement
            offset = ind_beg + len(section_replace_str)
                
        return res, locals_
    
    def __evaluate_function(self, fct_str, eval_params=EvaluateParameters.default_without_raise(), log_level=logging.DEBUG):
        func_name, args_str = self._get_function_and_args_strings(fct_str, with_interpreted=False)
        func = self._get_function(func_name) if func_name is not None else None
        if func is None:
            raise NotExistingFunctionTextInterpreterException("Cannot find a function in '{}'".format(fct_str))
        
        args = self._get_args(args_str)
        
        if Tools.do_log(logger, logging.TRACE, max_log_level=log_level):  # @UndefinedVariable
            logger.log(Tools.do_log_level(logging.TRACE, max_log_level=log_level), f"Evaluating function '{func_name}' with args {args}")  # @UndefinedVariable
            # logger.trace(f"Evaluating function '{func_name}' with args {args} (eval_params: {eval_params})", stack_info=True)
        try:
            res = func.apply(args)
        except BaseException as exc:
            if eval_params.raise_on_interpret_error:
                raise exc
            else:
                res = fct_str
                
        if Tools.do_log(logger, log_level):
            logger.log(log_level, f"Evaluate function [{fct_str}] => [{func_name}({', '.join(args)})] => [{res}] (type: {Typing.get_object_class_fullname(res)})")
        return res
    
    def __register_default_functions(self, dynamic_text_manager):
        self.register_function("len", FunctionApplyFunction(len, self, self._variable_manager))
        self.register_function("DynamicValue", FunctionDynamicValue(dynamic_text_manager))
        self.register_function("ExistsVariable", FunctionExistsVariable(self._variable_manager))
        
        # Convert
        self.register_function("int", FunctionCast(int, self, self._variable_manager))
        self.register_function("float", FunctionCast(float, self, self._variable_manager))
        self.register_function("str", FunctionCast(str, self, self._variable_manager))
        self.register_function("EscapeAllBytes", FunctionEscapeAllBytes(self, self._variable_manager))
        self.register_function("HexToBytes", FunctionHexToBytes(self._expression_evaluator))
        self.register_function("HexToInt", FunctionHexToInt(self, self._variable_manager))
        self.register_function("ToBase64", FunctionToBase64(self, self._variable_manager))
        self.register_function("ToBytes", FunctionToBytes(self, self._variable_manager))
        self.register_function("ToHex", FunctionToHex(self, self._variable_manager))
        self.register_function("ToString", FunctionToString(self, self._variable_manager))
        self.register_function("DatetimeUTCToTAI", FunctionConvert(lambda dt: DateTime.datetime_utc_to_tai(dt), self, self._variable_manager))
        self.register_function("DatetimeTAIToUTC", FunctionConvert(lambda dt: DateTime.datetime_tai_to_utc(dt), self, self._variable_manager))
        
    
    
    
    



# -*- coding: utf-8 -*-

#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################


from holado_test.scenario.step_tools import StepTools
from holado.common.context.session_context import SessionContext
from holado_test.behave.behave import *  # @UnusedWildImport
from holado_core.common.exceptions.functional_exception import FunctionalException
from holado_core.common.exceptions.technical_exception import TechnicalException
from holado_scripting.common.tools.variable_manager import VariableManager
from holado_test.common.context.scenario_context import ScenarioContext
from holado_s3.tools.s3.boto3_s3_client import Boto3S3Client
from holado_s3.tools.s3.minio_client import MinioS3Client
import logging
from holado_core.common.exceptions.verify_exception import VerifyException
from holado_value.common.tables.converters.value_table_converter import ValueTableConverter
from holado_test.behave.scenario.behave_step_tools import BehaveStepTools
from holado_scripting.text.verifier.text_verifier import TextVerifier
from holado_python.standard_library.typing import Typing

logger = logging.getLogger(__name__)


def __get_session_context() -> SessionContext:
    return SessionContext.instance()

def __get_scenario_context() -> ScenarioContext:
    return __get_session_context().get_scenario_context()

def __get_text_verifier() -> TextVerifier:
    return __get_scenario_context().get_text_verifier()

def __get_variable_manager() -> VariableManager:
    return __get_scenario_context().get_variable_manager()

def __get_path_manager():
    return SessionContext.instance().path_manager


################################################"
# Define steps specific to Minio S3 client
################################################"

if MinioS3Client.is_available():

    @Given(r"(?P<var_name>{Variable}) = new Minio S3 client")
    def step_impl(context, var_name):
        var_name = StepTools.evaluate_variable_name(var_name)
        table = BehaveStepTools.convert_step_table_2_value_table_with_header(context.table)
        kwargs = ValueTableConverter.convert_name_value_table_2_dict(table)
        
        try:
            client = MinioS3Client(**kwargs)
        except Exception as exc:
            raise FunctionalException(str(exc)) from exc

        __get_variable_manager().register_variable(var_name, client)
        
    @Step(r"execute \[(?P<command>{Any})\] \(Minio S3 client: (?P<var_client>{Variable})\)")
    def step_impl(context, command, var_client):
        """
        Execute any method in class Minio (cf https://min.io/docs/minio/linux/developers/python/API.html).
        An exception is raised if the method returns a result.
        """
        # command = StepTools.evaluate_scenario_parameter(command)
        command = StepTools.evaluate_string_parameter(command)
        client = StepTools.evaluate_variable_value(var_client)  # @UnusedVariable
        
        result = eval("client.internal_client." + command)
        if result:
            raise TechnicalException(f"Unexpected result to command [{command}]: {result}")
        
    @Step(r"(?P<var_name>{Variable}) = result of \[(?P<command>{Any})\] \(Minio S3 client: (?P<var_client>{Variable})\)")
    def step_impl(context, var_name, command, var_client):
        """
        Execute any method in class Minio (cf https://docs.min.io/docs/python-client-api-reference).
        The result is stored in given variable.
        """
        var_name = StepTools.evaluate_variable_name(var_name)
        # command = StepTools.evaluate_scenario_parameter(command)
        command = StepTools.evaluate_string_parameter(command)
        client = StepTools.evaluate_variable_value(var_client)  # @UnusedVariable
        
        result = eval("client.internal_client." + command)
        
        __get_variable_manager().register_variable(var_name, result)
        


################################################"
# Define steps specific to Boto3 S3 client
################################################"

if Boto3S3Client.is_available():

    @Given(r"(?P<var_name>{Variable}) = new Boto3 S3 client")
    def step_impl(context, var_name):
        var_name = StepTools.evaluate_variable_name(var_name)
        table = BehaveStepTools.convert_step_table_2_value_table_with_header(context.table)
        kwargs = ValueTableConverter.convert_name_value_table_2_dict(table)

        try:
            client = Boto3S3Client(**kwargs)
        except Exception as exc:
            raise FunctionalException(str(exc)) from exc

        __get_variable_manager().register_variable(var_name, client)

    @Step(r"execute \[(?P<command>{Any})\] \(Boto3 S3 client: (?P<var_client>{Variable})\)")
    def step_impl(context, command, var_client):
        """
        Execute any method in class S3.Client (cf https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/s3.html).
        An exception is raised if the method returns a result.
        """
        # command = StepTools.evaluate_scenario_parameter(command)
        command = StepTools.evaluate_string_parameter(command)
        client = StepTools.evaluate_variable_value(var_client)  # @UnusedVariable

        result = eval("client.internal_client." + command)
        if result:
            raise TechnicalException(f"Unexpected result to command [{command}]: {result}")

    @Step(r"(?P<var_name>{Variable}) = result of \[(?P<command>{Any})\] \(Boto3 S3 client: (?P<var_client>{Variable})\)")
    def step_impl(context, var_name, command, var_client):
        """
        Execute any method in class S3.Client (cf https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/s3.html).
        The result is stored in given variable.
        """
        var_name = StepTools.evaluate_variable_name(var_name)
        # command = StepTools.evaluate_scenario_parameter(command)
        command = StepTools.evaluate_string_parameter(command)
        client = StepTools.evaluate_variable_value(var_client)  # @UnusedVariable

        result = eval("client.internal_client." + command)

        __get_variable_manager().register_variable(var_name, result)
        


################################################"
# Define steps generic for any S3 client
################################################"

if MinioS3Client.is_available() or Boto3S3Client.is_available():
    
    @Step(r"(?P<var_name>{Variable}) = list of buckets \(S3 client: (?P<var_client>{Variable})\)")
    def step_impl(context, var_name, var_client):
        var_name = StepTools.evaluate_variable_name(var_name)
        
        client = StepTools.evaluate_variable_value(var_client)
        if isinstance(client, MinioS3Client):
            execute_steps(u"""
                {keyword} __RESULT_LIST_BUCKETS__@ = result of [list_buckets()] (Minio S3 client: {var_client})
                """.format(keyword=context.step.keyword, var_client=var_client) )
            result = __get_variable_manager().get_variable_value("__RESULT_LIST_BUCKETS__@")
            res = [bucket.name for bucket in result]
            __get_variable_manager().register_variable(var_name, res)
        elif isinstance(client, Boto3S3Client):
            execute_steps(u"""
                {keyword} __RESULT_LIST_BUCKETS__@ = result of [list_buckets()] (Boto3 S3 client: {var_client})
                """.format(keyword=context.step.keyword, var_client=var_client) )
            result = __get_variable_manager().get_variable_value("__RESULT_LIST_BUCKETS__@")
            res = [rb['Name'] for rb in result['Buckets']]
            __get_variable_manager().register_variable(var_name, res)
        else:
            raise TechnicalException(f"Unmanaged S3 client type {Typing.get_object_class_fullname(client)}")
        
    @Step(r"create bucket (?P<bucket>{Str}) \(S3 client: (?P<var_client>{Variable})\)")
    def step_impl(context, bucket, var_client):
        bucket = StepTools.evaluate_scenario_parameter(bucket)
        
        client = StepTools.evaluate_variable_value(var_client)
        if isinstance(client, MinioS3Client):
            execute_steps(u"""
                {keyword} execute [make_bucket('{bucket}')] (Minio S3 client: {var_client})
                """.format(keyword=context.step.keyword, bucket=bucket, var_client=var_client) )
        elif isinstance(client, Boto3S3Client):
            execute_steps(u"""
                {keyword} __RESULT__@ = result of [create_bucket(Bucket='{bucket}')] (Boto3 S3 client: {var_client})
                Then __RESULT__@['ResponseMetadata']['HTTPStatusCode'] == 200
                """.format(keyword=context.step.keyword, bucket=bucket, var_client=var_client) )
        else:
            raise TechnicalException(f"Unmanaged S3 client type {Typing.get_object_class_fullname(client)}")
        
    @Step(r"(?P<var_name>{Variable}) = list of objects in bucket (?P<bucket>{Str}) \(S3 client: (?P<var_client>{Variable})\)")
    def step_impl(context, var_name, bucket, var_client):
        """Get list of objects from a bucket.
        An optional table can be added to filter objects (cf S3 client documentation for a list of possible filters).
        Example: filter on 'Prefix'
        """
        var_name = StepTools.evaluate_variable_name(var_name)
        bucket = StepTools.evaluate_scenario_parameter(bucket)
        client = StepTools.evaluate_variable_value(var_client)
        table = BehaveStepTools.get_step_table(context)
        
        # Build command parameters
        if table:
            params = ValueTableConverter.convert_table_with_header_to_dict(table)
        else:
            params = {}
        if isinstance(client, MinioS3Client):
            if 'Prefix' in params:
                params['prefix'] = params.pop('Prefix')
            if 'recursive' not in params:
                params['recursive'] = True
        elif isinstance(client, Boto3S3Client):
            if 'MaxKeys' not in params:
                params['MaxKeys'] = client.MAX_KEYS
            if 'Delimiter' not in params:
                params['Delimiter'] = client.FAKE_DELIMITER
        else:
            raise TechnicalException(f"Unmanaged S3 client type {Typing.get_object_class_fullname(client)}")
        
        # Call command
        if params:
            params_str = ", " + ", ".join([f"{k}='{v}'" if isinstance(v, str) else f"{k}={v}" for k, v in params.items()])
        else:
            params_str = ""
        if isinstance(client, MinioS3Client):
            execute_steps(u"""
                {keyword} __RESPONSE_LIST_OBJECTS__@ = result of [list_objects('{bucket}'{params})] (Minio S3 client: {var_client})
                {keyword} {var_name} = convert object value __RESPONSE_LIST_OBJECTS__@ to list
                """.format(keyword=context.step.keyword, var_name=var_name, bucket=bucket, params=params_str, var_client=var_client) )
        elif isinstance(client, Boto3S3Client):
            execute_steps(u"""
                {keyword} __RESPONSE_LIST_OBJECTS__@ = result of [list_objects_v2(Bucket='{bucket}'{params})] (Boto3 S3 client: {var_client})
                Then __RESPONSE_LIST_OBJECTS__@['ResponseMetadata']['HTTPStatusCode'] == 200
                Then __RESPONSE_LIST_OBJECTS__@['IsTruncated'] == False
                """.format(keyword=context.step.keyword, bucket=bucket, params=params_str, var_client=var_client) )
            result = __get_variable_manager().get_variable_value("__RESPONSE_LIST_OBJECTS__@")
            if 'Contents' in result:
                res = result['Contents']
            else:
                res = []
            __get_variable_manager().register_variable(var_name, res)
        else:
            raise TechnicalException(f"Unmanaged S3 client type {Typing.get_object_class_fullname(client)}")
        
    @Step(r"(?P<var_name>{Variable}) = list of object names in bucket (?P<bucket>{Str}) \(S3 client: (?P<var_client>{Variable})\)")
    def step_impl(context, var_name, bucket, var_client):
        var_name = StepTools.evaluate_variable_name(var_name)
        
        execute_steps(format_step_with_context(context, 
                           u"__LIST_OBJECTS__@ = list of objects in bucket {bucket} (S3 client: {var_client})"
                           .format(bucket=bucket, var_client=var_client) ) )
        list_obj = __get_variable_manager().get_variable_value("__LIST_OBJECTS__@")
        
        client = StepTools.evaluate_variable_value(var_client)
        if isinstance(client, MinioS3Client):
            res = [obj.object_name for obj in list_obj]
        elif isinstance(client, Boto3S3Client):
            res = [obj['Key'] for obj in list_obj]
        else:
            raise TechnicalException(f"Unmanaged S3 client type {Typing.get_object_class_fullname(client)}")
        __get_variable_manager().register_variable(var_name, res)
        
    @Step(r"(?P<var_name>{Variable}) = data of object (?P<obj_name>{Str}) in bucket (?P<bucket>{Str}) \(S3 client: (?P<var_client>{Variable})\)")
    def step_impl(context, var_name, obj_name, bucket, var_client):
        var_name = StepTools.evaluate_variable_name(var_name)
        obj_name = StepTools.evaluate_scenario_parameter(obj_name)
        bucket = StepTools.evaluate_scenario_parameter(bucket)
        
        client = StepTools.evaluate_variable_value(var_client)
        if isinstance(client, MinioS3Client):
            execute_steps(u"""
                {keyword} __RESPONSE_GET__@ = result of [get_object('{bucket}', '{obj_name}')] (Minio S3 client: {var_client})
                {keyword} {var_name} = __RESPONSE_GET__@.data
                {keyword} __TMP__@ = __RESPONSE_GET__@.close()
                {keyword} __TMP__@ = __RESPONSE_GET__@.release_conn()
                """.format(keyword=context.step.keyword, var_name=var_name, obj_name=obj_name, bucket=bucket, var_client=var_client) )
        elif isinstance(client, Boto3S3Client):
            execute_steps(u"""
                {keyword} __RESPONSE_GET__@ = result of [get_object(Bucket='{bucket}', Key='{obj_name}')] (Boto3 S3 client: {var_client})
                {keyword} {var_name} = __RESPONSE_GET__@['Body'].read()
                """.format(keyword=context.step.keyword, var_name=var_name, obj_name=obj_name, bucket=bucket, var_client=var_client) )
        else:
            raise TechnicalException(f"Unmanaged S3 client type {Typing.get_object_class_fullname(client)}")
        
    @Step(r"put data (?P<data>{Str}) in object (?P<obj_name>{Str}) in bucket (?P<bucket>{Str}) \(S3 client: (?P<var_client>{Variable})\)")
    def step_impl(context, data, obj_name, bucket, var_client):
        data = StepTools.evaluate_scenario_parameter(data)
        obj_name = StepTools.evaluate_scenario_parameter(obj_name)
        bucket = StepTools.evaluate_scenario_parameter(bucket)
        
        client = StepTools.evaluate_variable_value(var_client)
        if isinstance(client, MinioS3Client):
            raise TechnicalException(fr"This sentence is not possible with Minio S3 client ; use instead \"{context.step.keyword} put file {{file_path}} in object {{obj_name}} in bucket {{bucket}} \(S3 client: {{var_client}}\)\"")
        elif isinstance(client, Boto3S3Client):
            if not isinstance(data, bytes):
                raise FunctionalException(f"Unmanaged data type '{Typing.get_object_class_fullname(data)}' (possible type: bytes)")
            result = client.internal_client.put_object(Bucket=bucket, Key=obj_name, Body=data)
            if result['ResponseMetadata']['HTTPStatusCode'] != 200:
                raise TechnicalException(f"Failed to put data in object '{obj_name}' in bucket '{bucket}': {result['ResponseMetadata']}")
        else:
            raise TechnicalException(f"Unmanaged S3 client type {Typing.get_object_class_fullname(client)}")
        
    @Step(r"get file (?P<file_path>{Str}) from object (?P<obj_name>{Str}) in bucket (?P<bucket>{Str}) \(S3 client: (?P<var_client>{Variable})\)")
    def step_impl(context, file_path, obj_name, bucket, var_client):
        file_path = StepTools.evaluate_scenario_parameter(file_path)
        obj_name = StepTools.evaluate_scenario_parameter(obj_name)
        bucket = StepTools.evaluate_scenario_parameter(bucket)
        
        __get_path_manager().makedirs(file_path)
        
        client = StepTools.evaluate_variable_value(var_client)
        if isinstance(client, MinioS3Client):
            execute_steps(u"""
                {keyword} __STAT__@ = result of [fget_object('{bucket}', '{obj_name}', '{file_path}')] (Minio S3 client: {var_client})
                """.format(keyword=context.step.keyword, file_path=file_path, obj_name=obj_name, bucket=bucket, var_client=var_client) )
        elif isinstance(client, Boto3S3Client):
            execute_steps(u"""
                {keyword} execute [download_file(Bucket='{bucket}', Key='{obj_name}', Filename='{file_path}')] (Boto3 S3 client: {var_client})
                """.format(keyword=context.step.keyword, file_path=file_path, obj_name=obj_name, bucket=bucket, var_client=var_client) )
        else:
            raise TechnicalException(f"Unmanaged S3 client type {Typing.get_object_class_fullname(client)}")
        
    @Step(r"put file (?P<file_path>{Str}) in object (?P<obj_name>{Str}) in bucket (?P<bucket>{Str}) \(S3 client: (?P<var_client>{Variable})\)")
    def step_impl(context, file_path, obj_name, bucket, var_client):
        file_path = StepTools.evaluate_scenario_parameter(file_path)
        obj_name = StepTools.evaluate_scenario_parameter(obj_name)
        bucket = StepTools.evaluate_scenario_parameter(bucket)
        
        client = StepTools.evaluate_variable_value(var_client)
        if isinstance(client, MinioS3Client):
            execute_steps(u"""
                {keyword} __WRITE_RESULT__@ = result of [fput_object('{bucket}', '{obj_name}', '{file_path}')] (Minio S3 client: {var_client})
                """.format(keyword=context.step.keyword, file_path=file_path, obj_name=obj_name, bucket=bucket, var_client=var_client) )
        elif isinstance(client, Boto3S3Client):
            with open(file_path, 'rb') as fin:
                file_content = fin.read()
            __get_variable_manager().register_variable("__BYTES__@", file_content)
            execute_steps(u"""
                {keyword} put data __BYTES__@ in object '{obj_name}' in bucket '{bucket}' (S3 client: {var_client})
                """.format(keyword=context.step.keyword, obj_name=obj_name, bucket=bucket, var_client=var_client) )
        else:
            raise TechnicalException(f"Unmanaged S3 client type {Typing.get_object_class_fullname(client)}")
    
    @Then(r"object named (?P<obj_name>{Str}) exists in bucket (?P<bucket>{Str}) \(S3 client: (?P<var_client>{Variable})\)")
    def step_impl(context, obj_name, bucket, var_client):
        obj_name = StepTools.evaluate_scenario_parameter(obj_name)
        
        client = StepTools.evaluate_variable_value(var_client)
        if isinstance(client, MinioS3Client):
            execute_steps(u"""
                Given __LIST_RESULT__@ = list of object names in bucket {bucket} (S3 client: {var_client})
                    | prefix     |
                    | '{prefix}' |
                """.format(bucket=bucket, var_client=var_client, prefix=obj_name) )
        elif isinstance(client, Boto3S3Client):
            execute_steps(u"""
                Given __LIST_RESULT__@ = list of object names in bucket {bucket} (S3 client: {var_client})
                    | Prefix     |
                    | '{prefix}' |
                """.format(bucket=bucket, var_client=var_client, prefix=obj_name) )
        else:
            raise TechnicalException(f"Unmanaged S3 client type {Typing.get_object_class_fullname(client)}")
        list_result = __get_variable_manager().get_variable_value("__LIST_RESULT__@")
        
        found = False
        for list_obj in list_result:
            found = __get_text_verifier().verify(list_obj, obj_name, raise_exception=False)
            if found:
                break
            
        if not found:
            raise VerifyException(f"Object '{obj_name}' doesn't exist in bucket '{StepTools.evaluate_scenario_parameter(bucket, log_level=logging.NOTSET)}' (existing {len(list_result)} objects: {list_result})")
    
    @Then(r"object named with prefix (?P<obj_name_prefix>{Str}) exists in bucket (?P<bucket>{Str}) \(S3 client: (?P<var_client>{Variable})\)")
    def step_impl(context, obj_name_prefix, bucket, var_client):
        obj_name_prefix = StepTools.evaluate_scenario_parameter(obj_name_prefix)
        
        client = StepTools.evaluate_variable_value(var_client)
        if isinstance(client, MinioS3Client):
            execute_steps(u"""
                Given __LIST_RESULT__@ = list of object names in bucket {bucket} (S3 client: {var_client})
                    | prefix     |
                    | '{prefix}' |
                """.format(bucket=bucket, var_client=var_client, prefix=obj_name_prefix) )
        elif isinstance(client, Boto3S3Client):
            execute_steps(u"""
                Given __LIST_RESULT__@ = list of object names in bucket {bucket} (S3 client: {var_client})
                    | Prefix     |
                    | '{prefix}' |
                """.format(bucket=bucket, var_client=var_client, prefix=obj_name_prefix) )
        else:
            raise TechnicalException(f"Unmanaged S3 client type {Typing.get_object_class_fullname(client)}")
        list_result = __get_variable_manager().get_variable_value("__LIST_RESULT__@")
        
        found = len(list_result) > 0
        if not found:
            raise VerifyException(f"Object with prefix '{obj_name_prefix}' doesn't exist in bucket '{StepTools.evaluate_scenario_parameter(bucket, log_level=logging.NOTSET)}' (existing {len(list_result)} objects: {list_result})")
  
